Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.

Section Arith.

  (* ============================================== *)
  (* EXPRESSIONS                                    *)
  (* ============================================== *)

  Inductive Arith (a : Set) : Set :=
  | Lit : nat -> Arith a
  | Add : a -> a -> Arith a.

  Definition Arith_fmap (B C : Set) (f : B -> C) (Aa : Arith B) : Arith C :=
    match Aa with
      | Lit n => Lit _ n
      | Add a b => Add _ (f a) (f b)
    end.

  Global Instance Arith_Functor : Functor Arith :=
    {| fmap := Arith_fmap |}.
    destruct a; reflexivity.
    (* fmap id *)
    destruct a; reflexivity.
  Defined.

  Variable F : Set -> Set.
  Context {Fun_F : Functor F}.
  Definition Exp := Exp F.
  Context {Sub_Arith_F : Arith :<: F}.

  (* Constructor + Universal Property. *)
   Context {WF_Sub_Arith_F : WF_Functor _ _ Sub_Arith_F _ _}.
  Definition lit' (n : nat) : Exp := inject' (Lit _ n).
  Definition lit  (n : nat) : Fix F := proj1_sig (lit' n).
  Global Instance UP'_lit {n : nat} :
    Universal_Property'_fold (lit n) := proj2_sig (lit' n).

  Definition add' (m n : Exp) : Exp :=  inject' (Add _ m n).

  Definition add (m n : Fix F)
    {UP'_m : Universal_Property'_fold m}
    {UP'_n : Universal_Property'_fold n}
    : Fix F := proj1_sig (add' (exist _ _ UP'_m) (exist _ _ UP'_n)).

   Global Instance UP'_add {m n : Fix F}
     {UP'_m : Universal_Property'_fold m}
     {UP'_n : Universal_Property'_fold n}
     :
     Universal_Property'_fold (add m n) :=
     proj2_sig (add' (exist _ _ UP'_m) (exist _ _ UP'_n)).

  (* Induction Principles for Arith. *)
  Definition ind_alg_Arith
    (P : forall e : Fix F, Universal_Property'_fold e -> Prop)
    (H : forall n, UP'_P P (lit n))
    (H0 : forall m n
      (IHm : UP'_P P m)
      (IHn : UP'_P P n),
      UP'_P P (@add m n (proj1_sig IHm) (proj1_sig IHn)))
      (e : Arith (sig (UP'_P P)))
        :
        sig (UP'_P P) :=
    match e with
      | Lit n => exist _ (lit n) (H n)
      | Add m n => exist (UP'_P P) _
        (H0 (proj1_sig m) (proj1_sig n) (proj2_sig m) (proj2_sig n))
    end.

  Definition ind2_alg_Arith
    {E E' : Set -> Set}
    {Fun_E : Functor E}
    {Fun_E' : Functor E'}
    {Sub_Arith_E : Arith :<: E}
    {Sub_Arith_E' : Arith :<: E'}
    (P : forall e : (Fix E) * (Fix E'),
      Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e) -> Prop)
    (H : forall n, UP'_P2 P (inject (Lit _ n), inject (Lit _ n)))
    (H0 : forall m n
      (IHm : UP'_P2 P m)
      (IHn : UP'_P2 P n),
      UP'_P2 P (inject (Add _ (exist _ _ (proj1 (proj1_sig IHm)))
        (exist _ _ (proj1 (proj1_sig IHn)))),
      inject (Add _ (exist _ _ (proj2 (proj1_sig IHm)))
        (exist _ _ (proj2 (proj1_sig IHn))))))
      (e : Arith (sig (UP'_P2 P)))
        :
        sig (UP'_P2 P) :=
    match e with
      | Lit n => exist _ _ (H n)
      | Add m n => exist (UP'_P2 P) _
        (H0 (proj1_sig m) (proj1_sig n) (proj2_sig m) (proj2_sig n))
    end.

  Lemma ind_PAlg_Arith (Name : Set)
    (P : forall e : Fix F, Universal_Property'_fold e -> Prop)
    (H : forall n, UP'_P P (lit n))
    (H0 : forall m n
      (IHm : UP'_P P m)
      (IHn : UP'_P P n),
      UP'_P P (@add m n (proj1_sig IHm) (proj1_sig IHn)))
    :
    PAlgebra Name Arith F (UP'_P P).
    econstructor 1 with (p_algebra := ind_alg_Arith P H H0).
    destruct e; simpl.
    unfold lit, lit'; simpl; rewrite wf_functor; reflexivity.
    unfold add, add'; simpl; rewrite wf_functor; reflexivity.
  Qed.

  Lemma ind_P2Alg_Arith (Name : Set)
    (E E' : Set -> Set)
    {Fun_E : Functor E}
    {Fun_E' : Functor E'}
    {Sub_Arith_E : Arith :<: E}
    {Sub_Arith_E' : Arith :<: E'}
    {WF_Fun_E : WF_Functor _ _ _ _ Fun_E}
    {WF_Fun_E' : WF_Functor _ _ _ _ Fun_E'}
    (P : forall e : (Fix E) * (Fix E'),
      Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e) -> Prop)
    (H : forall n, UP'_P2 P (inject (Lit _ n), inject (Lit _ n)))
    (H0 : forall m n
      (IHm : UP'_P2 P m)
      (IHn : UP'_P2 P n),
      UP'_P2 P (inject (Add _ (exist _ _ (proj1 (proj1_sig IHm)))
        (exist _ _ (proj1 (proj1_sig IHn)))),
      inject (Add _ (exist _ _ (proj2 (proj1_sig IHm)))
        (exist _ _ (proj2 (proj1_sig IHn)))))) :
    P2Algebra Name Arith E E' (UP'_P2 P).
    econstructor 1 with (p2_algebra := ind2_alg_Arith P H H0);
    destruct e; simpl; unfold inject; simpl; rewrite wf_functor; reflexivity.
  Qed.

  (* Alternative Induction Principles for Arith. *)
  Definition indD_alg_Arith
    (P : forall e : Fix F, Universal_Property'_fold e -> Set)
    (H : forall n, UP'_SP P (lit n))
    (H0 : forall m n
      (IHm : UP'_SP P m)
      (IHn : UP'_SP P n),
      UP'_SP P (@add m n (projT1 IHm) (projT1 IHn)))
      (e : Arith (sigT (UP'_SP P)))
        :
        sigT (UP'_SP P) :=
    match e with
      | Lit n => existT _ (lit n) (H n)
      | Add m n => existT (UP'_SP P) _
        (H0 (projT1 m) (projT1 n) (projT2 m) (projT2 n))
    end.

  Lemma ind_DAlg_Arith (Name : Set)
    (P : forall e : Fix F, Universal_Property'_fold e -> Set)
    (H : forall n, UP'_SP P (lit n))
    (H0 : forall m n
      (IHm : UP'_SP P m)
      (IHn : UP'_SP P n),
      UP'_SP P (@add m n (projT1 IHm) (projT1 IHn)))
    :
    DAlgebra Name Arith F (UP'_SP P).
    econstructor 1 with (d_algebra := indD_alg_Arith P H H0).
    destruct e; simpl.
    unfold lit, lit'; simpl; rewrite wf_functor; reflexivity.
    unfold add, add'; simpl; rewrite wf_functor; reflexivity.
  Qed.

  (* ============================================== *)
  (* VALUES                                         *)
  (* ============================================== *)

  (* Arithmetic Values. *)
   Inductive NatValue (A : Set) : Set :=
   | VI : nat -> NatValue A.

   Definition VI_fmap : forall (A B : Set) (f : A -> B),
     NatValue A -> NatValue B :=
     fun A B _ e => match e with
                      | VI n => VI _ n
                    end.

   Global Instance VI_Functor : Functor NatValue :=
     {| fmap := VI_fmap |}.
     destruct a; reflexivity.
     destruct a; reflexivity.
   Defined.

   Variable V : Set -> Set.
   Context {Fun_V : Functor V}.
   Definition Value := Value V.
   Context {Sub_NatValue_V : NatValue :<: V}.

  (* Constructor + Universal Property. *)
  Context {WF_Sub_NatValue_F : WF_Functor _ _ Sub_NatValue_V _ _}.

  Definition vi' (n : nat) : Value := inject' (VI _ n).
  Definition vi (n : nat) : Fix V := proj1_sig (vi' n).

   Global Instance UP'_vi {n : nat} :
     Universal_Property'_fold (vi n) := proj2_sig (vi' n).

   (* Constructor Testing for Arithmetic Values. *)

   Definition isVI : Fix V -> option nat :=
     fun exp =>
       match project exp with
        | Some (VI n) => Some n
        | None        => None
       end.

    Lemma project_vi_vi : forall n,
      project (F := V) (G := NatValue) (vi n) = Some (VI _ n).
    Proof.
      intros; unfold project, vi, inject; simpl; rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl; unfold VI_fmap.
      rewrite prj_inj; reflexivity.
    Qed.

    Lemma project_vi'_vi : forall n,
      project (F := V) (G := NatValue) (proj1_sig (vi' n)) = Some (VI _ n).
    Proof.
      intros; unfold project, vi', inject; simpl; rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl; unfold VI_fmap.
      rewrite prj_inj; reflexivity.
    Qed.

   Context {Sub_StuckValue_V : StuckValue :<: V}.
   Definition stuck' : nat -> Value := stuck' _.
   Definition stuck : nat -> Fix V := stuck _.

  (* ============================================== *)
  (* EVALUATION                                     *)
  (* ============================================== *)

   Context {Sub_BotValue_V : BotValue :<: V}.

   (* Evaluation Algebra for Arithemetic Expressions. *)

   Definition Arith_eval (R : Set) (rec : R -> evalR V)
     (e : Arith R) : evalR V :=
       match e with
         | Lit n => (fun _ => vi' n)
         | Add m n => (fun env =>
           let m' := (rec m env) in
             let n' := (rec n env) in
               match isVI (proj1_sig m'), isVI (proj1_sig n') with
                 | Some m', Some n' => vi' (m' + n')
                 | _, _ =>
                   if @isBot _ Fun_V Sub_BotValue_V (proj1_sig m') then @bot' _ Fun_V Sub_BotValue_V else
                     if @isBot _ Fun_V Sub_BotValue_V (proj1_sig n') then @bot' _ Fun_V Sub_BotValue_V else
                       stuck' 4
               end)
       end.

  Global Instance MAlgebra_eval_Arith T :
    FAlgebra EvalName T (evalR V) Arith :=
    {| f_algebra := Arith_eval T |}.

  Variable (ME : Set -> Set). (* Evaluation Monad. *)
  Context {Fun_M : Functor ME}.
  Context {Mon_M : Monad ME}.
  Context {Fail_M : FailMonad ME}.

  (* Monadic Evaluation Algebra for Arithemetic Expressions. *)

  Definition Arith_evalM (R : Set) (rec : R -> evalMR V ME)
     (e : Arith R) : evalMR V ME :=
       match e with
         | Lit n => return_ (vi' n)
         | Add m n => do m' <- rec m;
                      do n' <- rec n;
                        return_
                        match isVI (proj1_sig m'), isVI (proj1_sig n') with
                          | Some m', Some n' => vi' (m' + n')
                          | _, _ => stuck' 4
                        end
       end.

  Global Instance MAlgebra_evalM_Arith T :
    FAlgebra EvalName T (evalMR V ME) Arith :=
    {| f_algebra := Arith_evalM T|}.

  (* ============================================== *)
  (* PRETTY PRINTING                                *)
  (* ============================================== *)

  (* Pretty Printing Functions*)

  Require Import Ascii.
  Require Import String.

  Global Instance MAlgebra_ExpPrint_Arith T :
    FAlgebra ExpPrintName T (ExpPrintR) Arith :=
    {| f_algebra := fun rec e => match e with
      | Lit n => fun i => String (ascii_of_nat (n + 48)) EmptyString
      | Add m n => fun i => append "(" ((rec m i) ++ " + " ++ (rec n i) ++ ")")
    end |}.

  Global Instance MAlgebra_ValuePrint_VI T :
    FAlgebra ValuePrintName T ValuePrintR NatValue :=
    {| f_algebra := fun rec e =>
      match e with
        | VI n => String (ascii_of_nat (n + 48)) EmptyString
      end |}.

  (* ============================================== *)
  (* CONTINUITY Of MONADIC EVALUATION               *)
  (* ============================================== *)

  Context {evalM_F : FAlgebra EvalName Exp (evalMR V ME) F}.
  Context {WF_eval_F : @WF_FAlgebra EvalName _ _ Arith F
    Sub_Arith_F (MAlgebra_evalM_Arith _) evalM_F}.

   (* Continuity of Monadic Evaluation. *)

  Variable SV : (SubValue_i V -> Prop) -> SubValue_i V -> Prop.
  Context {Sub_SV_refl_SV : Sub_iFunctor (SubValue_refl V) SV}.
  Variable SVM : (SubValueM_i V ME -> Prop) -> SubValueM_i V ME -> Prop.
  Context {iFun_SVM : iFunctor SVM}.
  Context {Sub_SVM_base_SVM : Sub_iFunctor (SubValueM_base V ME) SVM}.

  (* Lit case. *)

  Ltac WF_FAlg_rewrite := repeat rewrite wf_functor; simpl;
    repeat rewrite out_in_fmap; simpl;
      repeat rewrite wf_functor; simpl;
        repeat rewrite wf_algebra; simpl.

    Lemma bevalM_continuous_Exp_H : forall n,
      UP'_SP (bevalM_continuous_Exp_P V F ME SVM) (lit n).
      unfold bevalM_continuous_Exp_P; intros; econstructor; intros.
      exact UP'_lit.
      unfold bevalM, mfold, lit; simpl.
      simpl; rewrite H1;
        unfold bevalM, mfold, lit; simpl; WF_FAlg_rewrite.
      apply inject_i; constructor.
    Qed.

    Context {SVM_Bind_SVM: iPAlgebra SVM_Bind_Name (SVM_Bind_P V ME SVM) SVM}.

    (* Definition Arith_eval_continuous_Addn_P (i : SubValueM_i V ME) := *)
    (*   SubValueM V ME SVM i /\ *)
    (*   forall v, SubValueMC V ME SVM  *)
    (*     (do n' <- svm_a V ME i; *)
    (*       return_ *)
    (*       match isVI v with *)
    (*         | Some m'0 => *)
    (*           match isVI (proj1_sig n') with *)
    (*             | Some n'0 => vi' (m'0 + n'0) *)
    (*             | None => stuck' 4 *)
    (*           end *)
    (*         | None => stuck' 4 *)
    (*       end) *)
    (*     (do n' <- svm_b V ME i; *)
    (*       return_ *)
    (*       match isVI v with *)
    (*         | Some m'0 => *)
    (*           match isVI (proj1_sig n') with *)
    (*             | Some n'0 => vi' (m'0 + n'0) *)
    (*             | None => stuck' 4 *)
    (*           end *)
    (*         | None => stuck' 4 *)
    (*       end). *)

    (* Inductive Arith_eval_continuous_Addn_Name := AeCAn_name. *)

    (* Global Instance WFVM_base_eval_continuous_Addn : *)
    (*    iPAlgebra Arith_eval_continuous_Addn_Name Arith_eval_continuous_Addn_P (SubValueM_base V ME). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; apply ind_alg_SVM_base with (Fail_ME := Fail_M);  *)
    (*     try eassumption; unfold Arith_eval_continuous_Addn_P; split; simpl; intros. *)
    (*   (* SVM_Fail *) *)
    (*   apply inject_i; econstructor. *)
    (*   rewrite bind_fail; apply inject_i; constructor 1. *)
    (*   (* SVM_Return *) *)
    (*   apply inject_i; econstructor. *)
    (*   apply inject_i; constructor 2. *)
    (* Qed. *)

    (* Definition Arith_eval_continuous_Addm_P (i : SubValueM_i V ME) := *)
    (*   SubValueM V ME SVM i /\ *)
    (*   forall n1 n2, *)
    (*     SubValueMC V ME SVM n1 n2 ->  *)
    (*     SubValueMC V ME SVM *)
    (*     (do m' <- svm_a V ME i; *)
    (*       do n' <- n1; *)
    (*         return_ match isVI (proj1_sig m') with *)
    (*                   | Some m'0 => *)
    (*                     match isVI (proj1_sig n') with *)
    (*                       | Some n'0 => vi' (m'0 + n'0) *)
    (*                       | None => stuck' 4 *)
    (*                     end *)
    (*                   | None => stuck' 4 *)
    (*                 end) *)
    (*     (do m' <- svm_b V ME i; *)
    (*       do n' <- n2; *)
    (*         return_ match isVI (proj1_sig m') with *)
    (*                   | Some m'0 => *)
    (*                     match isVI (proj1_sig n') with *)
    (*                       | Some n'0 => vi' (m'0 + n'0) *)
    (*                       | None => stuck' 4 *)
    (*                     end *)
    (*                   | None => stuck' 4 *)
    (*                 end). *)

    (* Inductive Arith_eval_continuous_Addm_Name := AecAm_name. *)

    (* Global Instance WFVM_base_eval_continuous_Addm : *)
    (*    iPAlgebra Arith_eval_continuous_Addm_Name Arith_eval_continuous_Addm_P (SubValueM_base V ME). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; apply ind_alg_SVM_base with (Fail_ME := Fail_M);  *)
    (*     try eassumption; unfold Arith_eval_continuous_Addm_P; split; simpl; intros. *)
    (*   (* SVM_Fail *) *)
    (*   apply inject_i; econstructor. *)
    (*   rewrite bind_fail; apply inject_i; constructor 1. *)
    (*   (* SVM_Return *) *)
    (*   apply inject_i; econstructor. *)
    (*   apply inject_i; constructor 2. *)
    (* Qed. *)

    Lemma bevalM_continuous_Exp_H0 : forall
      (m n : Fix F)
      (IHm : UP'_SP (bevalM_continuous_Exp_P V F ME SVM) m)
      (IHn : UP'_SP (bevalM_continuous_Exp_P V F ME SVM) n),
      UP'_SP (bevalM_continuous_Exp_P V F ME SVM) (@add m n (projT1 IHm) (projT1 IHn)).
      unfold bevalM_continuous_Exp_P; intros.
      destruct IHm as [m_UP' IHm].
      destruct IHn as [n_UP' IHn].
      econstructor; intros; eauto with typeclass_instances.
      unfold bevalM, mfold; simpl; rewrite H1; unfold add; simpl.
      WF_FAlg_rewrite.
      repeat erewrite bF_UP_in_out.
      generalize (H (exist Universal_Property'_fold m m_UP') _ _ H0 (refl_equal _))
        as SVM_m_m'; intros.
      generalize (H (exist Universal_Property'_fold n n_UP') _ _ H0 (refl_equal _))
        as SVM_n_n'; intros.
      apply SVM_Bind; eauto with typeclass_instances.
      intros; apply SVM_Bind; eauto with typeclass_instances.
      intros; apply inject_i; constructor; auto.
    Qed.

    Global Instance Arith_bevalM_continuous_Exp  :
      DAlgebra EC_ExpName Arith F (UP'_SP (bevalM_continuous_Exp_P V F ME SVM)).
    Proof.
      apply ind_DAlg_Arith.
      apply bevalM_continuous_Exp_H.
      apply bevalM_continuous_Exp_H0.
    Qed.

  (* ============================================== *)
  (* TYPES                                          *)
  (* ============================================== *)

  (* The Arithmetic Type. *)
  Inductive AType (A : Set) : Set :=
  | TNat : AType A.

  Definition AType_fmap : forall (A B : Set) (f : A -> B),
    AType A -> AType B := fun A B _ _ => TNat _.

  Global Instance AType_Functor : Functor AType :=
    {| fmap := AType_fmap |}.
    destruct a; reflexivity.
    (* fmap id *)
    destruct a; reflexivity.
  Defined.

  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.
  Definition DType := DType D.
  Context {Sub_AType_D  : AType :<: D}.

   (* Constructor + Universal Property. *)
  Context {WF_Sub_AType_D : WF_Functor _ _ Sub_AType_D _ _}.

  Definition tnat' : DType := inject' (TNat _).
  Definition tnat : Fix D := proj1_sig tnat'.
  Global Instance UP'_tnat :
    Universal_Property'_fold tnat := proj2_sig tnat'.

  (* Induction Principle for Nat Types. *)
  Definition ind_alg_AType 
    (P : forall d : Fix D, Universal_Property'_fold d -> Prop)
    (H : UP'_P P tnat)
    (d : AType (sig (UP'_P P))) : sig (UP'_P P) :=
    match d with
      | TNat => exist _ _ H
    end.

  Lemma ind_PAlg_AType {Sub_AType_D' : AType :<: D}
    (eq_Sub_AType : forall A (d : AType A),
      inj (Sub_Functor := Sub_AType_D) d = inj (Sub_Functor := Sub_AType_D') d)
    (Name : Set)
    (P : forall e : Fix D, Universal_Property'_fold e -> Prop)
    (H : UP'_P P tnat) :
    PAlgebra (sub_F_G := Sub_AType_D') Name AType D (UP'_P P).
    econstructor 1 with (p_algebra := ind_alg_AType P H).
    destruct e; simpl; unfold tnat, tnat'; simpl;
      rewrite wf_functor; apply f_equal; eauto.
  Qed.

  (* Type Equality Section. *)
  Definition isTNat : Fix D -> bool :=
   fun typ =>
     match project typ with
      | Some TNat => true
      | None      => false
     end.

    Lemma project_tnat_tnat : project (F := D) (G := AType) tnat = Some (TNat _).
    Proof.
      intros; unfold project, tnat, inject; simpl; rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl; unfold AType_fmap.
      rewrite prj_inj; reflexivity.
    Qed.

    Lemma project_tnat'_tnat :
      project (F := D) (G := AType) (proj1_sig tnat') = Some (TNat _).
    Proof.
      intros; unfold project, tnat', inject; simpl; rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl; unfold AType_fmap.
      rewrite prj_inj; reflexivity.
    Qed.

  Definition AType_eq_DType  (R : Set) (rec : R -> eq_DTypeR D)
    (e : AType R) : eq_DTypeR D :=
    match e with
      | TNat => fun t => isTNat (proj1_sig t)
    end.

  Global Instance MAlgebra_eq_DType_Arith T:
    FAlgebra eq_DTypeName T (eq_DTypeR D) AType :=
    {| f_algebra := AType_eq_DType T|}.

  Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.
  Context {WF_DType_eq_DT : forall T, @WF_FAlgebra eq_DTypeName T _ _ _
    Sub_AType_D (MAlgebra_eq_DType_Arith T) (eq_DType_DT _)}.

  Global Instance PAlgebra_eq_DType_eq_AType {Sub_AType_D' : AType :<: D}
    (eq_Sub_AType : forall A (d : AType A),
      inj (Sub_Functor := Sub_AType_D) d = inj (Sub_Functor := Sub_AType_D') d)
    :
    PAlgebra (sub_F_G := Sub_AType_D') eq_DType_eqName AType D (UP'_P (eq_DType_eq_P D)).
  Proof.
    apply ind_PAlg_AType; eauto.
    unfold UP'_P; econstructor.
    unfold eq_DType_eq_P; intros.
    unfold eq_DType, mfold, tnat, tnat', inject' in H; simpl in H;
      repeat rewrite wf_functor in H; simpl in H; unfold in_t in H.
    rewrite (wf_algebra (WF_FAlgebra := WF_DType_eq_DT _)) in H; simpl in H.
    unfold isTNat in H.
    caseEq (project (proj1_sig d2)); rewrite H0 in H;
      try discriminate; destruct a.
    unfold project in H0.
    apply inj_prj in H0.
    unfold tnat, tnat'; simpl; rewrite wf_functor; simpl.
    unfold AType_fmap.
    generalize (f_equal (in_t_UP' _ _ ) H0); intros.
    apply (f_equal (@proj1_sig _ _)) in H1;
      rewrite in_out_UP'_inverse in H1.
    rewrite H1; simpl; rewrite wf_functor; simpl; unfold AType_fmap;
      reflexivity.
    exact (proj2_sig d2).
  Qed.
  Global Instance PAlgebra_eq_DType_neq_AType 
    {Sub_AType_D' : AType :<: D}
    (eq_Sub_AType : forall A (d : AType A),
      inj (Sub_Functor := Sub_AType_D) d = inj (Sub_Functor := Sub_AType_D') d)
    :
    PAlgebra (sub_F_G := Sub_AType_D') eq_DType_neqName AType D (UP'_P (eq_DType_neq_P D)).
  Proof.
    apply ind_PAlg_AType; eauto.
    unfold UP'_P; econstructor.
    unfold eq_DType_neq_P; intros.
    unfold eq_DType, mfold, tnat, tnat', inject' in H; simpl in H;
      repeat rewrite wf_functor in H; simpl in H; unfold in_t in H.
    rewrite (wf_algebra (WF_FAlgebra := WF_DType_eq_DT _)) in H; simpl in H.
    unfold isTNat in H.
    caseEq (project (proj1_sig d2)); rewrite H0 in H;
      try (destruct a; discriminate).
    unfold not; intros; rewrite <- H1 in H0.
    unfold project, tnat, inject in H0; simpl in H0.
    rewrite out_in_fmap in H0; repeat rewrite wf_functor in H0;
      simpl in H0; rewrite prj_inj in H0; discriminate.
  Qed.

  (* ============================================== *)
  (* TYPING                                         *)
  (* ============================================== *)

  (* Typing Arithmetic Expressions. *)

  Variable (MT : Set -> Set).
  Context {Fun_MT : Functor MT}.
  Context {Mon_MT : Monad MT}.
  Context {Fail_MT : FailMonad MT}.

  Definition Arith_typeof (R : Set) (rec : R -> typeofR D MT)
     (e : Arith R) : typeofR D MT :=
     match e with
       | Lit n => return_ tnat'
       | Add m n => do tm <- rec m;
                    do tn <- rec n;
                      match isTNat (proj1_sig tm), isTNat (proj1_sig tn) with
                          | true, true => return_ tnat'
                          | _, _ => fail
                        end
     end.

  Global Instance MAlgebra_typeof_Arith T:
    FAlgebra TypeofName T (typeofR D MT) Arith :=
    {| f_algebra := Arith_typeof T|}.

  (* ============================================== *)
  (* WELL-FORMED NAT VALUES                         *)
  (* ============================================== *)

  Variable TypContext : Set.

  Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
  Variable funWFV : iFunctor WFV.

    (** Natural Numbers are well-formed **)

    Inductive WFValue_VI (WFV : WFValue_i D V TypContext -> Prop) : WFValue_i D V TypContext -> Prop :=
    | WFV_VI : forall Sigma n v T,
      proj1_sig v = vi n ->
      proj1_sig T = tnat ->
      WFValue_VI WFV (mk_WFValue_i D V _ Sigma v T).

    Definition ind_alg_WFV_VI (P : WFValue_i D V TypContext -> Prop)
      (H : forall Sigma n v T veq Teq, P (mk_WFValue_i _ _ _ Sigma v T))
      i (e : WFValue_VI P i) : P i :=
      match e in WFValue_VI _ i return P i with
        | WFV_VI Sigma n v T veq Teq => H Sigma n v T veq Teq
      end.

    Definition WFV_VI_ifmap (A B : WFValue_i D V TypContext -> Prop) i (f : forall i, A i -> B i)
      (WFV_a : WFValue_VI A i) : WFValue_VI B i :=
      match WFV_a in (WFValue_VI _ s) return (WFValue_VI B s)
        with
        | WFV_VI Sigma n v T veq Teq => WFV_VI B Sigma n v T veq Teq
      end.

    Global Instance iFun_WFV_VI : iFunctor WFValue_VI.
      constructor 1 with (ifmap := WFV_VI_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable Sub_WFV_VI_WFV : Sub_iFunctor WFValue_VI WFV.

    Global Instance WFV_proj1_a_VI :
      iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V TypContext WFV) WFValue_VI.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFV_proj1_a_P.
      inversion H; subst; simpl; intros.
      apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; simpl; eauto.
      rewrite H3; eauto.
    Defined.

     Global Instance WFV_proj1_b_VI :
       iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V TypContext WFV) WFValue_VI.
       econstructor; intros.
       unfold iAlgebra; intros; unfold WFV_proj1_b_P.
       inversion H; subst; simpl; intros.
       apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; simpl; eauto.
       rewrite H3; eauto.
     Defined.

    (* Inversion principles for Well-formed natural numbers. *)
    Definition WF_invertVI_P (i : WFValue_i D V TypContext) :=
      proj1_sig (wfv_T _ _ _ i) = tnat ->
      WFValue_VI (iFix WFV) i.

    Inductive WF_invertVI_Name := wfv_invertvi_name.
    Context {WF_invertVI_WFV :
      iPAlgebra WF_invertVI_Name WF_invertVI_P WFV}.

    Global Instance WF_invertVI_VI :
      iPAlgebra WF_invertVI_Name WF_invertVI_P WFValue_VI.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertVI_P.
      inversion H; subst; simpl; intros.
      econstructor; eassumption.
    Defined.

    Definition WF_invertVI := ifold_ WFV _ (ip_algebra (iPAlgebra := WF_invertVI_WFV)).

    Context {TypContextCE : ConsExtensionC TypContext}.

    Global Instance WFV_Weaken_VI :
      iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V TypContext WFV) WFValue_VI.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFValue_Weaken_P.
      inversion H; subst; simpl; intros.
      apply inject_i; econstructor; eassumption.
    Defined.

    Context {WFV_proj1_a_WFV :
      iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V _ WFV) WFV}.
    Context {WFV_proj1_b_WFV :
       iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV) WFV}.

    Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) F}.
    Context {WF_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ Arith F
      Sub_Arith_F (MAlgebra_typeof_Arith _) (Typeof_F T)}.

    Context {evalM_F' : forall T, FAlgebra EvalName T (evalMR V ME) F}.
    Context {WF_evalM_F' : forall T, @WF_FAlgebra EvalName _ _ Arith F
      Sub_Arith_F (MAlgebra_evalM_Arith _) (evalM_F' T)}.

    Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
    Context {funWFVM : iFunctor WFVM}.

    Context {Sub_WFVM_base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME _ WFV) WFVM}.

    Lemma Arith_eval_Soundness_H :
      forall n : nat,
        UP'_SP (eval_soundness_P D V F MT ME _ WFVM) (lit n).
    Proof.
      intros; econstructor.
      unfold eval_soundness_P; intros; simpl.
      unfold typeof, evalM, mfold, lit; simpl; unfold in_t; simpl.
      WF_FAlg_rewrite.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor 1 with (T := tnat').
      apply inject_i; econstructor; unfold vi; simpl; auto.
      apply return_orn; auto.
    Defined.

    (* Sub-Lemma for soundness of add *)

    Definition Arith_eval_Soundness_Addn_P (i : WFValueM_i D V MT ME _) :=
      WFValueM D V MT ME _ WFVM i /\
      forall v T (WF_v_T : WFValueC D V _ WFV (wfvm_S D V MT ME _ i) v T)
        (mt' : MT {T' : Names.DType D | proj1_sig T = proj1_sig T'}),
        WFValueMC D V MT ME _ WFVM (wfvm_S D V MT ME _ i)
        (do n' <- wfvm_v D V MT ME _ i;
          return_
          match isVI (proj1_sig v) with
            | Some m'0 =>
              match isVI (proj1_sig n') with
                | Some n'0 => vi' (m'0 + n'0)
                | None => stuck' 4
              end
            | None => stuck' 4
          end)
        (do tm <- fmap (@proj1_sig _ _) mt';
          do tn <- wfvm_T D V MT ME _ i;
            (if isTNat (proj1_sig tm)
              then if isTNat (proj1_sig tn) then return_ tnat' else fail
              else fail)).

    Inductive Arith_eval_Soundness_Addn_Name := AeSAn_name.

    Global Instance WFVM_base_eval_Soundness_Addn :
       iPAlgebra Arith_eval_Soundness_Addn_Name Arith_eval_Soundness_Addn_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_base with (WFV := WFV);
        try eassumption; unfold Arith_eval_Soundness_Addn_P; simpl; split; intros.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor; eauto.
      rename WF_v_T into WF_v_T0; rename H0 into WF_v_T.
      simpl; repeat rewrite <- left_unit.
      unfold isTNat, isVI.
      destruct H1 as [mt'' mt''_eq]; subst.
      caseEq (project (proj1_sig T0)).
      destruct a.
      caseEq (project (proj1_sig T)).
      destruct a.
      rename H0 into T0_eq; rename H1 into T_eq.
      apply project_inject in T_eq; apply project_inject in T0_eq;
        auto with typeclass_instances; try exact (proj2_sig _).
      generalize (WF_invertVI _ WF_v_T T_eq) as WF_v_T';
      generalize (WF_invertVI _ WF_v_T0 T0_eq) as WF_v_T0'; simpl; intros.
      inversion WF_v_T0'; inversion WF_v_T'; subst.
      rewrite H2, H7, project_vi_vi, project_vi_vi.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor 1 with (T := tnat').
      apply inject_i; econstructor.
      unfold vi; simpl; auto.
      unfold tnat; simpl; auto.
      apply bind_orn; intros.
      destruct a as [T' T'_eq]; subst; simpl.
      apply bind_orn; intros.
      destruct a as [T0' T0'_eq]; subst; simpl.
      rewrite <- T'_eq, H4, project_tnat_tnat, <- T0'_eq, H9, project_tnat_tnat; simpl.
      apply return_orn; reflexivity.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)).
      eapply WFVM_Untyped3_T with (mt' := fmap (@proj1_sig _ _) mt'');
        intros T1 T1_eq; rewrite <- T1_eq, H0; auto.
      rewrite <- eq_bind_fmap with
      (f := (fun tn : {T' : Names.DType D | proj1_sig T = proj1_sig T'} =>
        if match project (proj1_sig (proj1_sig tn)) with
             | Some TNat => true
             | None => false
           end
          then return_ tnat'
          else fail)); auto.
      rewrite fmap_m; unfold wbind; rewrite <- associativity;
        apply rewrite_do; apply functional_extensionality; intros.
      rewrite <- left_unit; destruct x as [T' T'_eq]; simpl; rewrite <- T'_eq, H1; auto.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)).
      eapply WFVM_Untyped3_T with (mt' := fmap (@proj1_sig _ _) mt''); intros.
      intros; simpl; unfold wbind; apply f_equal;
        apply functional_extensionality; intros; rewrite <- H1, H0; auto.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor; eauto.
      simpl; unfold wbind.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM));
        eapply WFVM_Untyped4_T; intros.
      rewrite <- associativity, bind_fail; reflexivity.
    Qed.

    Context {WFVM_eval_Soundness_Addn :
      iPAlgebra Arith_eval_Soundness_Addn_Name Arith_eval_Soundness_Addn_P WFVM}.

    Definition eval_Soundness_Addn :=
      ifold_ WFVM _ (ip_algebra (iPAlgebra := WFVM_eval_Soundness_Addn)).

    Definition Arith_eval_Soundness_Addm_P (i : WFValueM_i D V MT ME _) :=
      WFValueM D V MT ME _ WFVM i /\
      forall n1 n2,
        (forall Sigma, ConsExtension Sigma (wfvm_S D V MT ME _ i) ->
          WFValueMC D V MT ME _ WFVM Sigma n1 n2) -> 
        WFValueMC D V MT ME _ WFVM (wfvm_S D V MT ME _ i)
        (do m' <- wfvm_v D V MT ME _ i;
          do n' <- n1;
            return_ match isVI (proj1_sig m') with
                      | Some m'0 =>
                        match isVI (proj1_sig n') with
                          | Some n'0 => vi' (m'0 + n'0)
                          | None => stuck' 4
                        end
                      | None => stuck' 4
                    end)
        (do tm <- wfvm_T D V MT ME _ i;
          do tn <- n2;
            (if isTNat (proj1_sig tm)
              then if isTNat (proj1_sig tn) then return_ tnat' else fail
              else fail)).

    Inductive Arith_eval_Soundness_Addm_Name := AeSAm_name.

    Global Instance WFVM_base_eval_Soundness_Addm :
       iPAlgebra Arith_eval_Soundness_Addm_Name Arith_eval_Soundness_Addm_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_base with (WFV := WFV);
        try assumption; unfold Arith_eval_Soundness_Addm_P; intros; split; intros.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor; eauto.
      simpl; repeat rewrite <- left_unit.
      destruct H1 as [mt' mt'_eq]; subst.
      apply ((proj2 (eval_Soundness_Addn _ (H2 Sigma (ConsExtension_id _)))) _ _ H0).
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor 2.
      simpl; unfold wbind; rewrite <- associativity, bind_fail.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); econstructor 2.
    Qed.

    Context {WFVM_eval_Soundness_Addm :
      iPAlgebra Arith_eval_Soundness_Addm_Name Arith_eval_Soundness_Addm_P WFVM}.

    Definition eval_Soundness_Addm :=
      ifold_ WFVM _ (ip_algebra (iPAlgebra := WFVM_eval_Soundness_Addm)).

    Lemma Arith_eval_soundness'_H0 
     (typeof_R eval_R : Set) typeof_rec eval_rec : 
     forall (m1 n1 : typeof_R) (m2 n2 : eval_R)
       (Sigma : TypContext)
       (IHm1 :  WFValueMC D V MT ME _ WFVM Sigma
         (eval_rec m2) (typeof_rec m1))
       (IHm2 : forall Sigma', ConsExtension Sigma' Sigma -> 
         WFValueMC D V MT ME _ WFVM Sigma'
         (eval_rec n2) (typeof_rec n1)),
       WFValueMC D V MT ME _ WFVM Sigma
         (Arith_evalM _ eval_rec (Add _ m2 n2))
         (Arith_typeof _ typeof_rec (Add _ m1 n1)).
    Proof.
      simpl; intros.
      apply ((proj2 (eval_Soundness_Addm _ IHm1)) _ _ IHm2).
    Qed.

    Section PSoundness.
      
      Variable P_bind : Set.
      Variable E' : Set -> Set.
      Context {Fun_E' : Functor E'}.
      Context {Sub_Arith_E' : Arith :<: E'}.
      Context {WF_Fun_E' : WF_Functor _ _ Sub_Arith_E' _ _}.

      Variable P : UP'_F F -> UP'_F E' -> P_bind -> TypContext -> Prop.
      
      Variable Add_Decomp1 : 
        forall m1 m2 n1 n2 m1_UP' m2_UP' m1_UP'' m2_UP'' add1_UP' add2_UP' pb Sigma, 
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Add _ (exist Universal_Property'_fold m2 m2_UP') n2)))) add2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Add _ (exist Universal_Property'_fold m1 m1_UP') n1)))) add1_UP')
          pb Sigma -> 
          P (exist _ m2 m2_UP'') (exist _ m1 m1_UP'') pb Sigma.

      Variable Add_Decomp2 : 
        forall m1 m2 n1 n2 n1_UP' n2_UP' n1_UP'' n2_UP'' add1_UP' add2_UP' pb Sigma, 
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Add _ m2 (exist Universal_Property'_fold n2 n2_UP'))))) add2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Add _ m1 (exist Universal_Property'_fold n1 n1_UP'))))) add1_UP')
          pb Sigma -> 
          P (exist _ n2 n2_UP'') (exist _ n1 n1_UP'') pb Sigma.

    Global Instance Arith_eval_soundness'

      (P_Sub : forall e e' pb Sigma Sigma', 
        P e e' pb Sigma -> ConsExtension Sigma' Sigma -> 
        P e e' pb Sigma')
      {Typeof_E' : forall T, FAlgebra TypeofName T (typeofR _ MT) E'}
      {WF_typeof_E' : forall T, @WF_FAlgebra TypeofName T _ _ _
        Sub_Arith_E' (MAlgebra_typeof_Arith T) (Typeof_E' _)}
      (pb : P_bind)
      (eval_rec : Exp -> evalMR V ME)
      (typeof_rec : UP'_F E' -> typeofR D MT)
      :
      P2Algebra ES'_ExpName Arith E' F 
      (UP'_P2 (eval_soundness'_P D V F MT ME _ WFVM
        _ _ _ P pb typeof_rec eval_rec (f_algebra (FAlgebra := Typeof_E' _ ))
        (f_algebra (FAlgebra := evalM_F)))).
    Proof.
      apply ind_P2Alg_Arith; eauto.
      (* lit n case *)
      unfold eval_soundness'_P, UP'_P2; simpl; intros;
        constructor 1 with 
          (x := conj (proj2_sig (inject' (Lit _ n))) (proj2_sig (inject' (Lit _ n)))).
      unfold inject; simpl; repeat rewrite out_in_fmap;
        repeat rewrite wf_functor; intros.
      WF_FAlg_rewrite.
      apply (inject_i (subGF := Sub_WFVM_base_WFVM)); 
        constructor 1 with (T := tnat').
      apply inject_i; econstructor; unfold vi; simpl; eauto.
      apply return_orn; eauto.
      (* add case *)
      destruct m as [m1 m2]; destruct n as [n1 n2];
        destruct IHm as [[UP'_m1 UP'_m2] IHm];
          destruct IHn as [[UP'_n1 UP'_n2] IHn]; simpl in *|-*.
      assert (Universal_Property'_fold
    (fst
       (inject
          (Add (sig Universal_Property'_fold)
             (exist Universal_Property'_fold m1 (proj1 (conj UP'_m1 UP'_m2)))
             (exist Universal_Property'_fold n1 (proj1 (conj UP'_n1 UP'_n2)))),
       inject
         (Add (sig Universal_Property'_fold)
            (exist Universal_Property'_fold m2 (proj2 (conj UP'_m1 UP'_m2)))
            (exist Universal_Property'_fold n2 (proj2 (conj UP'_n1 UP'_n2)))))) /\
  Universal_Property'_fold
    (snd
       (inject
          (Add (sig Universal_Property'_fold)
             (exist Universal_Property'_fold m1 (proj1 (conj UP'_m1 UP'_m2)))
             (exist Universal_Property'_fold n1 (proj1 (conj UP'_n1 UP'_n2)))),
       inject
         (Add (sig Universal_Property'_fold)
            (exist Universal_Property'_fold m2 (proj2 (conj UP'_m1 UP'_m2)))
            (exist Universal_Property'_fold n2 (proj2 (conj UP'_n1 UP'_n2))))))).
      simpl; split; unfold inject; exact (proj2_sig _).
      unfold eval_soundness'_P, UP'_P2; intros.
      constructor 1 with (x := H).
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj Sigma WF_Sigma IHa.
      repeat rewrite wf_functor; simpl.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold Arith_fmap.
      intros; apply Arith_eval_soundness'_H0.
      intros; generalize (IHa _ _ (exist _ _ UP'_m1, exist _ _ UP'_m2) 
        (Add_Decomp1 _ _ _ _ _ _ _ _ _ _ _ _ WF_Sigma)); simpl.
      rewrite typeof_rec_proj; simpl; intros H'; apply H'.
      apply IHm; eauto.
      intros Sigma' ConsExt_Sig'_Sig;
        generalize (IHa _ _ (exist _ _ UP'_n1, exist _ _ UP'_n2)
          (Add_Decomp2 _ _ _ _ _ _ _ _ _ _ _ _ (P_Sub _ _ _ _ _
            WF_Sigma ConsExt_Sig'_Sig))); simpl.
      rewrite typeof_rec_proj; simpl; intros H'; apply H'.
      apply IHn; eauto.
    Defined.

  End PSoundness.

  Global Instance Arith_eval_soundness'' typeof_rec eval_rec :
    P2Algebra ES'_ExpName Arith F F
    (UP'_P2 (eval_soundness'_P D V F MT ME _ WFVM
      _ _ _ (fun _ _ _ _ => True) tt typeof_rec eval_rec
      (f_algebra (FAlgebra := Typeof_F _)) (f_algebra (FAlgebra := evalM_F)))).
  Proof.
    eapply Arith_eval_soundness'; eauto.
  Defined.

    (* Add case.

    Definition bevalM_continuous_Arith_P (i : SubValueM_i V ME) :=
      forall (m0 n0 : nat) (e : Exp),
        SubValueMC V ME SVM (bevalM _ _ _ m0 e) (bevalM _ _ _ n0 e) ->
        SubValueMC V ME SVM
        (do m' <- svm_a _ _ i;
          if isBot V (proj1_sig m') then (return_ (bot' V)) else
            do n' <- bevalM V F ME (bevalM_E := evalM_F) m0 e;
              return_
              match isVI (proj1_sig m'), isVI (proj1_sig n') with
                | Some m', Some n' => vi' (m' + n')
                | _, _ => if isBot V (proj1_sig n')
                  then bot' V
                  else stuck' 4
              end)
        (do m' <- svm_b V ME i;
          if isBot V (proj1_sig m') then (return_ (bot' V)) else
            do n' <- bevalM V F ME (bevalM_E := evalM_F) n0 e;
              return_
              match isVI (proj1_sig m'), isVI (proj1_sig n') with
                | Some m', Some n' => vi' (m' + n')
                | _, _ => if isBot V (proj1_sig n')
                  then bot' V
                  else stuck' 4
              end).

    Definition bevalM_continuous_Arith_Return_P (i : SubValueM_i V ME) :=
      forall (a b : Names.Value V),
        SubValueC V SV a b ->
        isBot V (proj1_sig a) = false ->
        SubValueMC V ME SVM
        (do n' <- (svm_a V ME i);
          return_
          match isVI (proj1_sig a), isVI (proj1_sig n') with
            | Some m', Some n' => vi' (m' + n')
            | _, _ => if isBot V (proj1_sig n')
              then bot' V
              else stuck' 4
          end)
        (do n' <- (svm_b V ME i);
          return_
          match isVI (proj1_sig b), isVI (proj1_sig n') with
            | Some m', Some n' => vi' (m' + n')
            | _, _ => if isBot V (proj1_sig n')
              then bot' V
              else stuck' 4
          end).

    Inductive bevalM_continuous_Arith_Return_Name := bevalmc_arith_Return_name.

    Global Instance Value_bevalM_continuous_Arith_Value  :
      iPAlgebra bevalM_continuous_Arith_Return_Name bevalM_continuous_Arith_Return_P (SubValueM_Return V ME SV).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_Return_P;
        intros; inversion H; subst; simpl.
      repeat rewrite <- left_unit.
      apply (inject_i (subGF := Sub_SubValueM_Return_SVM)); econstructor.
      unfold isVI.
      caseEq (project (G := NatValue) (proj1_sig a)).
      apply project_inject in H3; auto with typeclass_instances.
      destruct n.
      generalize (SV_invertVI _ H0 n H3); simpl; intro prj_svb; rewrite prj_svb.
      caseEq (project (G := NatValue) (proj1_sig a0)); auto.
      destruct n0.
      apply project_inject in H4; auto with typeclass_instances.
      generalize (SV_invertVI _ H2 _ H4); simpl; intro prj_svb0; rewrite prj_svb0.
      repeat rewrite project_vi_vi; apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; eauto.
      exact (proj2_sig _).
      unfold isBot; caseEq (project (G := BotValue) (proj1_sig a0)); try destruct b1; simpl.
      apply inject_i; econstructor; eauto.
      rewrite project_vi_vi.
      caseEq (project (G := NatValue) (proj1_sig b0)).
      destruct n0.
      apply project_inject in H6; auto with typeclass_instances.
      destruct (SV_invertVI' _ H2 _ H6) as [n_eq_vi | n_eq_bot];
        simpl in *|-.
      rewrite n_eq_vi in H4.
      unfold vi, project in H4; simpl in H4; rewrite out_in_fmap in H4;
        repeat rewrite wf_functor in H4; rewrite prj_inj in H4;
          discriminate.
      rewrite n_eq_bot in H5.
      unfold bot, project in H5; simpl in H5; rewrite out_in_fmap in H5;
        repeat rewrite wf_functor in H5; rewrite prj_inj in H5;
          discriminate.
      exact (proj2_sig _).
      caseEq (project (G := BotValue) (proj1_sig b0)); try destruct b1; simpl.
      apply project_inject in H7; auto with typeclass_instances.
      generalize (SV_invertBot _ SV _ _ H2 H7); simpl; intro a0_eq;
        rewrite a0_eq in H5.
      rewrite project_bot_bot in H5; discriminate.
      exact (proj2_sig _).
      apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; eauto.
      exact (proj2_sig _).
      unfold isBot; caseEq (project (G := BotValue) (proj1_sig a0));
        try destruct b1.
      apply inject_i; econstructor; eauto.
      caseEq (project (G := NatValue) (proj1_sig b)).
      destruct n.
      apply project_inject in H5; auto with typeclass_instances.
      destruct (SV_invertVI' _ H0 _ H5) as [n_eq_vi | n_eq_bot];
        simpl in *|-.
      rewrite n_eq_vi in H3.
      unfold vi, project in H3; simpl in H3; rewrite out_in_fmap in H3;
        repeat rewrite wf_functor in H3; rewrite prj_inj in H3;
          discriminate.
      rewrite n_eq_bot in H1.
      unfold isBot in H1; rewrite project_bot_bot in H1; discriminate.
      exact (proj2_sig _).
      caseEq (project (G := BotValue) (proj1_sig b0)).
      destruct b1.
      apply project_inject in H6; auto with typeclass_instances.
      generalize (SV_invertBot _ SV _ _ H2 H6); simpl; intro b0_eq;
        rewrite b0_eq in H4; rewrite project_bot_bot in H4; discriminate.
      exact (proj2_sig _).
      apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; eauto.
    Qed.

    Global Instance Value_bevalM_continuous_Bot_Value  :
      iPAlgebra bevalM_continuous_Arith_Return_Name bevalM_continuous_Arith_Return_P (SubValueM_Bot V ME).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_Return_P;
        intros; inversion H; subst; simpl.
      repeat rewrite <- left_unit.
      unfold isVI.
      caseEq (project (G := NatValue) (proj1_sig a)).
      apply project_inject in H2; auto with typeclass_instances.
      destruct n.
      generalize project_vi_bot as H3; unfold bot; simpl; intro;
        rewrite H3.
      generalize project_bot_bot as H4; unfold isBot, bot; simpl;
        intro; rewrite H4.
      apply inject_i; econstructor.
      exact (proj2_sig _).
      generalize project_bot_bot as H4; unfold isBot, bot; simpl;
        intro; rewrite H4.
      apply inject_i; econstructor.
    Qed.

    Context {bevalM_continuous_Arith_Return_SVM :
      iPAlgebra bevalM_continuous_Arith_Return_Name bevalM_continuous_Arith_Return_P SVM}.
    Definition bevalM_continuous_Arith_Value :=
      ifold_ SVM _ (ip_algebra (iPAlgebra := bevalM_continuous_Arith_Return_SVM)).

    Inductive bevalM_continuous_Arith_Name := bevalmc_arith_name.

    Global Instance Return_bevalM_continuous_Arith  :
      iPAlgebra bevalM_continuous_Arith_Name bevalM_continuous_Arith_P (SubValueM_Return V ME SV).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_P;
        intros; inversion H; subst; simpl.
      repeat rewrite <- left_unit.
      unfold isBot; caseEq (project (G := BotValue) (proj1_sig a)); try destruct b0; simpl.
      apply inject_i; econstructor; eauto.
      caseEq (project (G := BotValue) (proj1_sig b)); try destruct b0; simpl.
      apply project_inject in H3; auto with typeclass_instances.
      generalize (SV_invertBot _ SV _ _ H1 H3); simpl; intro a0_eq;
        rewrite a0_eq in H2.
      rewrite project_bot_bot in H2; discriminate.
      exact (proj2_sig _).
      apply (bevalM_continuous_Arith_Value _ H0 _ _ H1); simpl.
      unfold isBot; rewrite H2; auto.
    Qed.

    Global Instance Bot_bevalM_continuous_Arith  :
      iPAlgebra bevalM_continuous_Arith_Name bevalM_continuous_Arith_P (SubValueM_Bot V ME).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_P;
        intros; inversion H; subst; simpl.
      repeat rewrite <- left_unit.
      generalize project_bot_bot; unfold bot, isBot.
      intro eq_bot; rewrite eq_bot.
      apply inject_i; econstructor; eauto.
    Qed.

    Context {bevalM_continuous_Arith_SVM :
      iPAlgebra bevalM_continuous_Arith_Name bevalM_continuous_Arith_P SVM}.
    Definition bevalM_continuous_Arith :=
      ifold_ SVM _ (ip_algebra (iPAlgebra := bevalM_continuous_Arith_SVM)).

    (* Global Instance Refl_bevalM_continuous_Arith_Refl  :  *)
    (*   iPAlgebra bevalM_continuous_Arith_Name bevalM_continuous_Arith_P (SubValueM_Value V ME SV). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_P; *)
    (*     intros; inversion H; subst; simpl. *)
    (*   apply (inject_i (subGF := Sub_SubValueM_Refl_SVM)); econstructor. *)
    (* Qed. *)

    (* Definition bevalM_continuous_Arith_Refl_P (i : SubValueM_i V ME) :=  *)
    (*   forall (m0 n0 : nat) (e : Exp) (Ma : ME Value), *)
    (*     SubValueMC V ME SVM (bevalM _ _ _ m0 e) (bevalM _ _ _ n0 e) ->  *)
    (*     SubValueMC V ME SVM *)
    (*     (do m' <- Ma; *)
    (*      do n' <-  svm_a _ _ i; *)
    (*        return_ match isVI (proj1_sig m') with *)
    (*        | Some m'0 => *)
    (*          match isVI (proj1_sig n') with *)
    (*            | Some n'0 => vi' (m'0 + n'0) *)
    (*            | None => *)
    (*              if isBot V (proj1_sig m') *)
    (*                then bot' V *)
    (*                else if isBot V (proj1_sig n') then bot' V else stuck' 4 *)
    (*          end *)
    (*        | None => *)
    (*          if isBot V (proj1_sig m') *)
    (*            then bot' V *)
    (*            else if isBot V (proj1_sig n') then bot' V else stuck' 4 *)
    (*      end) *)
    (*     (do m' <- Ma; *)
    (*      do n' <- svm_b V ME i; *)
    (*      return_ (M := ME) match isVI (proj1_sig m') with *)
    (*        | Some m'0 => *)
    (*          match isVI (proj1_sig n') with *)
    (*            | Some n'0 => vi' (m'0 + n'0) *)
    (*            | None => *)
    (*              if isBot V (proj1_sig m') *)
    (*                then bot' V *)
    (*                else if isBot V (proj1_sig n') then bot' V else stuck' 4 *)
    (*          end *)
    (*        | None => *)
    (*          if isBot V (proj1_sig m') *)
    (*            then bot' V *)
    (*            else if isBot V (proj1_sig n') then bot' V else stuck' 4 *)
    (*      end). *)

    (* Inductive bevalM_continuous_Arith_Refl_Name := bevalmc_arith_refl_name. *)

    (* Global Instance Refl_bevalM_continuous_Arith_Refl  :  *)
    (*   iPAlgebra bevalM_continuous_Arith_Refl_Name bevalM_continuous_Arith_Refl_P (SubValueM_Refl V ME). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_Refl_P; *)
    (*     intros; inversion H; subst; simpl. *)
    (*   apply (inject_i (subGF := Sub_SubValueM_Refl_SVM)); econstructor. *)
    (* Qed. *)

    (* Global Instance Value_bevalM_continuous_Arith_Refl  :  *)
    (*   iPAlgebra bevalM_continuous_Arith_Refl_Name bevalM_continuous_Arith_Refl_P (SubValueM_Value V ME SV). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_Refl_P; *)
    (*     intros; inversion H; subst; simpl. *)
    (*   apply (inject_i (subGF := Sub_SubValueM_Refl_SVM)); econstructor. *)
    (* Qed. *)


    (* Definition bevalM_continuous_Arith :=  *)
    (*   ifold_ SVM _ (ip_algebra (iPAlgebra := bevalM_continuous_Arith_SVM)). *)

    (* Global Instance Bot_bevalM_continuous_Arith  :  *)
    (*   iPAlgebra bevalM_continuous_Arith_Name bevalM_continuous_Arith_P (SubValueM_Refl V ME). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_P; *)
    (*     intros; inversion H; subst; simpl. *)
    (*   rewrite <- left_unit. *)
    (*   generalize project_vi_bot; unfold bot, isVI, bot'. simpl.  *)
    (*   intro H2; rewrite H2. *)
    (*   generalize project_bot_bot; unfold bot, isBot; simpl; intro H3; *)
    (*     rewrite H3. *)
    (*   apply (inject_i (subGF := Sub_SubValueM_Value_SVM)); econstructor. *)
    (*   apply inject *)

    (*   generalize (bevalM_continuous_Arith_Bot _ H0). *)
    (*   unfold bevalM_continuous_Arith_Bot_P, bot', isVI, isBot. *)
    (*   intro H4; apply H4. *)
    (* Qed. *)

    (* Global Instance bevalM_continuous_Arith_Refl  :  *)
    (*   iPAlgebra bevalM_continuous_Arith_Name bevalM_continuous_Arith_P (SubValueM_Refl V ME). *)
    (* Proof. *)
    (*   econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith_P; *)
    (*     intros; inversion H; subst; simpl. *)

    Lemma bevalM_continuous_Exp_H0 : forall
      (m n : Fix F)
      (IHm : UP'_SP (bevalM_continuous_Exp_P V F ME SVM) m)
      (IHn : UP'_SP (bevalM_continuous_Exp_P V F ME SVM) n),
      UP'_SP (bevalM_continuous_Exp_P V F ME SVM) (@add m n (projT1 IHm) (projT1 IHn)).
      unfold bevalM_continuous_Exp_P; intros.
      destruct IHm as [m_UP' IHm].
      destruct IHn as [n_UP' IHn].
      econstructor; intros; eauto with typeclass_instances.
      destruct n0.
      elimtype False; inversion H0.
      assert (m0 <= n0) as le_m0_n0 by auto with arith; clear H0.
      unfold bevalM, mfold; simpl; rewrite H1; unfold add; simpl.
      WF_FAlg_rewrite.
      repeat erewrite bF_UP_in_out.
      generalize (H (exist Universal_Property'_fold m m_UP') _ _ le_m0_n0 (refl_equal _))
        as SVM_m_m'; intros.
      generalize (H (exist Universal_Property'_fold n n_UP') _ _ le_m0_n0 (refl_equal _))
        as SVM_n_n'; intros.
      unfold SubValueMC in *.
      eapply (bevalM_continuous_Arith _ SVM_m_m'); eauto.
    Qed. *)

    (* Lemma bevalM_continuous_Exp'_H : forall n,
      UP'_SP (bevalM_continuous_Exp'_P V F ME SV) (lit n).
      unfold bevalM_continuous_Exp'_P; intros; econstructor; intros.
      exact UP'_lit.
      unfold bevalM, mfold, lit; simpl.
      simpl; rewrite H0;
        unfold bevalM, mfold, lit; simpl; WF_FAlg_rewrite.
      assert (SubValueC V SV (bot' V) (vi' n)) by (apply inject_i; econstructor; eauto).
      unfold SubValueC in H1; exists (return_ (exist _ _ H1)); simpl;
        repeat rewrite fmap_m, <- left_unit; split; reflexivity.
    Qed.

    (* Definition bevalM_continuous_Arith'_P (i : SubValueM_i V ME) :=
      forall (n0 : nat) (e : Exp),
        svm_a _ _ i = return_ (bot' V) ->
        SubValueMC V ME SVM (return_ (bot' V)) (bevalM _ _ _ n0 e) ->
        SubValueMC V ME SVM (return_ (bot' V))
        (do m' <- svm_b V ME i;
         do n' <- bevalM V F ME (bevalM_E := evalM_F) n0 e;
         return_ (M := ME) match isVI (proj1_sig m') with
           | Some m'0 =>
             match isVI (proj1_sig n') with
               | Some n'0 => vi' (m'0 + n'0)
               | None =>
                 if isBot V (proj1_sig m')
                   then bot' V
                   else if isBot V (proj1_sig n') then bot' V else stuck' 4
             end
           | None =>
             if isBot V (proj1_sig m')
               then bot' V
               else if isBot V (proj1_sig n') then bot' V else stuck' 4
         end).

    Inductive bevalM_continuous_Arith'_Name := bevalmc_arith'_name.

    Definition bevalM_continuous_Arith'_Value_P (i : SubValueM_i V ME) :=
      forall (Mb : ME (Names.Value V)),
        {Mm : Product ME ME {i : SubValue_i V | SubValue V SV i} |
          fst (fmap (fun sv_i : {i : SubValue_i V | SubValue V SV i} =>
            sv_a V (proj1_sig sv_i)) Mm) = return_ (bot' V) /\
          snd (fmap (fun sv_i : {i : SubValue_i V | SubValue V SV i} =>
            sv_b V (proj1_sig sv_i)) Mm) = Mb} ->
        svm_a _ _ i = return_ (bot' V) ->
        SubValueMC V ME SVM (return_ (bot' V))
        (do m' <- Mb;
          do n' <- (svm_b V ME i);
            return_
            match isVI (proj1_sig m') with
              | Some m'0 =>
                match isVI (proj1_sig n') with
                  | Some n'0 => vi' (m'0 + n'0)
                  | None =>
                    if isBot V (proj1_sig m')
                      then bot' V
                      else if isBot V (proj1_sig n') then bot' V else stuck' 4
                end
              | None =>
                if isBot V (proj1_sig m')
                  then bot' V
                  else if isBot V (proj1_sig n') then bot' V else stuck' 4
            end).

    Lemma fmap_do_return : forall (M : Set -> Set) `{Monad M} (A B C : Set)
      (ma : M A) (f : A -> B) (g : B -> C),
      fmap g (do x <- ma; return_ (f x)) = do x <- ma; return_ (g (f x)).
    Proof.
      intros; rewrite fmap_m, <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      rewrite <- left_unit; reflexivity.
    Qed.

    Lemma fmap_do_return2 : forall (M : Set -> Set) `{Monad M} (A B C D : Set)
      (ma : M A) (mb : M B) (f : A -> B -> C) (g : C -> D),
      fmap g (do x <- ma; do y <- mb; return_ (f x y)) =
      do x <- ma; do y <- mb; return_ (g (f x y)).
    Proof.
      intros; rewrite fmap_m, <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      intros; repeat rewrite <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      rewrite <- left_unit; reflexivity.
    Qed.

    Lemma fmap_dist : forall (M : Set -> Set) `{Monad M} (A B C : Set)
      (ma : M A) (f : A -> B) (g : B -> M C),
      do x <- ma; g (f x) = do x <- (fmap f ma); g x.
    Proof.
      intros; rewrite fmap_m, <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      rewrite <- left_unit; reflexivity.
    Qed.

    Inductive bevalM_continuous_Arith'_Value_Name := bevalmc_arith'_value_name.

    Global Instance Value_bevalM_continuous_Arith'_Value  :
      iPAlgebra bevalM_continuous_Arith'_Value_Name bevalM_continuous_Arith'_Value_P (SubValueM_Value V ME SV).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith'_Value_P;
        intros; inversion H; subst; simpl.
      destruct H0 as [m' [m_eq m_eq']]; destruct H2 as [n' [n_eq n_eq']].
      apply (inject_i (subGF := Sub_SubValueM_Value_SVM)); econstructor.
      generalize bevalM_continuous_Add as H0'; intros;  unfold SubValueC in H0'.
      exists (do m <- m'; do n <- n'; return_ (M := Product ME ME) (exist _ _ (H0' m n))).
      destruct m' as [xm xn]; destruct n' as [xm0 xn0]; simpl in *|-*.
      repeat rewrite fmap_do_return2; simpl; split; intros.
      rewrite fmap_dist with (ma := xm) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_a V (proj1_sig x)))
        (g := fun x => do y <- xm0;
          return_
          match isVI (proj1_sig x) with
            | Some m'0 =>
              match isVI (proj1_sig (sv_a V (proj1_sig y))) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig x)
                    then bot' V
                    else
                      if isBot V (proj1_sig (sv_a V (proj1_sig y)))
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig x)
                then bot' V
                else
                  if isBot V (proj1_sig (sv_a V (proj1_sig y)))
                    then bot' V
                    else stuck' 4
          end).
      rewrite m_eq, <- left_unit.
      rewrite fmap_dist with (ma := xm0) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_a V (proj1_sig x)))
        (g := fun y => return_
          match isVI (proj1_sig (bot' V)) with
            | Some m'0 =>
              match isVI (proj1_sig y) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig (bot' V))
                    then bot' V
                    else
                      if isBot V (proj1_sig y)
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig (bot' V))
                then bot' V
                else
                  if isBot V (proj1_sig y)
                    then bot' V
                    else stuck' 4
          end).
      rewrite n_eq, H1, <- left_unit.
      unfold isVI; generalize project_vi_bot; unfold bot;
        intro pr_v_b; rewrite pr_v_b.
      unfold isBot; generalize project_bot_bot; unfold bot;
        intro pr_b_b; rewrite pr_b_b; auto.
      rewrite fmap_dist with (ma := xn) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_b V (proj1_sig x)))
        (g := fun x => do y <- xn0;
          return_
          match isVI (proj1_sig x) with
            | Some m'0 =>
              match isVI (proj1_sig (sv_b V (proj1_sig y))) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig x)
                    then bot' V
                    else
                      if isBot V (proj1_sig (sv_b V (proj1_sig y)))
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig x)
                then bot' V
                else
                  if isBot V (proj1_sig (sv_b V (proj1_sig y)))
                    then bot' V
                    else stuck' 4
          end).
      rewrite m_eq'; apply f_equal; apply functional_extensionality; intros.
      rewrite fmap_dist with (ma := xn0) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_b V (proj1_sig x)))
        (g := fun y => return_
          match isVI (proj1_sig x) with
            | Some m'0 =>
              match isVI (proj1_sig y) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig x)
                    then bot' V
                    else
                      if isBot V (proj1_sig y)
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig x)
                then bot' V
                else
                  if isBot V (proj1_sig y)
                    then bot' V
                    else stuck' 4
          end).
      rewrite n_eq'; reflexivity.
    Qed.

    Context {bevalM_continuous_Arith'_Value_SVM :
      iPAlgebra bevalM_continuous_Arith'_Value_Name bevalM_continuous_Arith'_Value_P SVM}.
    Definition bevalM_continuous_Arith'_Value :=
      ifold_ SVM _ (ip_algebra (iPAlgebra := bevalM_continuous_Arith'_Value_SVM)).

    Global Instance Value_bevalM_continuous_Arith'  :
      iPAlgebra bevalM_continuous_Arith'_Name bevalM_continuous_Arith'_P (SubValueM_Value V ME SV).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Arith'_P;
        intros; inversion H; subst; simpl.
      apply (bevalM_continuous_Arith'_Value _ H1); simpl; auto.
      destruct H2 as [Mm [fst_Mm snd_Mm]]; exists Mm; simpl;
        destruct Mm; simpl in *|-*; split; congruence.
    Qed.

    Context {bevalM_continuous_Arith'_SVM :
      iPAlgebra bevalM_continuous_Arith'_Name bevalM_continuous_Arith'_P SVM}.
    Definition bevalM_continuous_Arith' :=
      ifold_ SVM _ (ip_algebra (iPAlgebra := bevalM_continuous_Arith'_SVM)). *)

    Lemma fmap_do_return : forall (M : Set -> Set) `{Monad M} (A B C : Set)
      (ma : M A) (f : A -> B) (g : B -> C),
      fmap g (do x <- ma; return_ (f x)) = do x <- ma; return_ (g (f x)).
    Proof.
      intros; rewrite fmap_m, <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      rewrite <- left_unit; reflexivity.
    Qed.

    Lemma fmap_do_return2 : forall (M : Set -> Set) `{Monad M} (A B C D : Set)
      (ma : M A) (mb : M B) (f : A -> B -> C) (g : C -> D),
      fmap g (do x <- ma; do y <- mb; return_ (f x y)) =
      do x <- ma; do y <- mb; return_ (g (f x y)).
    Proof.
      intros; rewrite fmap_m, <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      intros; repeat rewrite <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      rewrite <- left_unit; reflexivity.
    Qed.

    Lemma fmap_dist : forall (M : Set -> Set) `{Monad M} (A B C : Set)
      (ma : M A) (f : A -> B) (g : B -> M C),
      do x <- ma; g (f x) = do x <- (fmap f ma); g x.
    Proof.
      intros; rewrite fmap_m, <- associativity.
      apply f_equal; apply functional_extensionality; intros.
      rewrite <- left_unit; reflexivity.
    Qed.

    Lemma bevalM_continuous_Exp'_H0 : forall
      (m n : Fix F)
      (IHm : UP'_SP (bevalM_continuous_Exp'_P V F ME SV) m)
      (IHn : UP'_SP (bevalM_continuous_Exp'_P V F ME SV) n),
      UP'_SP (bevalM_continuous_Exp'_P V F ME SV) (@add m n (projT1 IHm) (projT1 IHn)).
    Proof.
      unfold bevalM_continuous_Exp'_P; intros.
      destruct IHm as [m_UP' IHm].
      destruct IHn as [n_UP' IHn].
      econstructor; intros; eauto with typeclass_instances.
      unfold bevalM, mfold; simpl; rewrite H0; unfold add; simpl.
      WF_FAlg_rewrite.
      repeat erewrite bF_UP_in_out.
      generalize (H (exist Universal_Property'_fold m m_UP') _  (refl_equal _))
        as SVM_m_m'; intros.
      generalize (H (exist Universal_Property'_fold n n_UP') _ (refl_equal _))
        as SVM_n_n'; intros.
      destruct SVM_m_m' as [m' [m_eq m_eq']]; destruct SVM_n_n' as [n' [n_eq n_eq']].
      generalize bevalM_continuous_Add as H0'; intros;  unfold SubValueC in H0'.
      exists (do m <- m'; do n <- n'; return_ (M := Product ME ME) (exist _ _ (H0' m n))).
      destruct m' as [xm xn]; destruct n' as [xm0 xn0]; simpl in *|-*.
      repeat rewrite fmap_do_return2; simpl; split; intros.
      rewrite fmap_dist with (ma := xm) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_a V (proj1_sig x)))
        (g := fun x => do y <- xm0;
          return_
          match isVI (proj1_sig x) with
            | Some m'0 =>
              match isVI (proj1_sig (sv_a V (proj1_sig y))) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig x)
                    then bot' V
                    else
                      if isBot V (proj1_sig (sv_a V (proj1_sig y)))
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig x)
                then bot' V
                else
                  if isBot V (proj1_sig (sv_a V (proj1_sig y)))
                    then bot' V
                    else stuck' 4
          end).
      rewrite m_eq, <- left_unit.
      rewrite fmap_dist with (ma := xm0) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_a V (proj1_sig x)))
        (g := fun y => return_
          match isVI (proj1_sig (bot' V)) with
            | Some m'0 =>
              match isVI (proj1_sig y) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig (bot' V))
                    then bot' V
                    else
                      if isBot V (proj1_sig y)
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig (bot' V))
                then bot' V
                else
                  if isBot V (proj1_sig y)
                    then bot' V
                    else stuck' 4
          end).
      rewrite n_eq, <- left_unit.
      unfold isVI; generalize project_vi_bot; unfold bot;
        intro pr_v_b; rewrite pr_v_b.
      unfold isBot; generalize project_bot_bot; unfold bot;
        intro pr_b_b; rewrite pr_b_b; auto.
      rewrite fmap_dist with (ma := xn) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_b V (proj1_sig x)))
        (g := fun x => do y <- xn0;
          return_
          match isVI (proj1_sig x) with
            | Some m'0 =>
              match isVI (proj1_sig (sv_b V (proj1_sig y))) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig x)
                    then bot' V
                    else
                      if isBot V (proj1_sig (sv_b V (proj1_sig y)))
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig x)
                then bot' V
                else
                  if isBot V (proj1_sig (sv_b V (proj1_sig y)))
                    then bot' V
                    else stuck' 4
          end).
      rewrite m_eq'; apply f_equal; apply functional_extensionality; intros.
      rewrite fmap_dist with (ma := xn0) (f := fun x : {i : SubValue_i V | SubValue V SV i} => (sv_b V (proj1_sig x)))
        (g := fun y => return_
          match isVI (proj1_sig x) with
            | Some m'0 =>
              match isVI (proj1_sig y) with
                | Some n'0 => vi' (m'0 + n'0)
                | None =>
                  if isBot V (proj1_sig x)
                    then bot' V
                    else
                      if isBot V (proj1_sig y)
                        then bot' V
                        else stuck' 4
              end
            | None =>
              if isBot V (proj1_sig x)
                then bot' V
                else
                  if isBot V (proj1_sig y)
                    then bot' V
                    else stuck' 4
          end).
      rewrite n_eq'; reflexivity.
    Qed.

    Global Instance Arith_bevalM_continuous_Exp'  :
      DAlgebra EC_ExpName' Arith F (UP'_SP (bevalM_continuous_Exp'_P V F ME SV)).
    Proof.
      apply ind_DAlg_Arith.
      apply bevalM_continuous_Exp'_H.
      apply bevalM_continuous_Exp'_H0.
    Qed. *)

    (* Lemma evalM_continuous_Exp_H0 : forall  *)
    (*   (m n : Fix F) *)
    (*   (IHm : UP'_P (evalM_continuous_Exp_P V F ME SV) m)  *)
    (*   (IHn : UP'_P (evalM_continuous_Exp_P V F ME SV) n),  *)
    (*   UP'_P (evalM_continuous_Exp_P V F ME SV) (@add m n (proj1_sig IHm) (proj1_sig IHn)). *)
    (*   unfold evalM_continuous_Exp_P; intros. *)
    (*   destruct IHm as [m_UP' IHm]. *)
    (*   destruct IHn as [n_UP' IHn]. *)
    (*   econstructor; intros; eauto with typeclass_instances. *)
    (*   unfold bevalM, mfold, add; simpl. *)
    (*   WF_FAlg_rewrite. *)
    (*   repeat erewrite bF_UP_in_out. *)
    (*   destruct (H (exist Universal_Property'_fold m m_UP')). *)
    (*   destruct (H (exist Universal_Property'_fold n n_UP')). *)
    (*   unfold bevalM in H0, H1. *)
    (*   rewrite <- H0, <- H1. *)
    (*   assert (forall (m' n' : Names.Value V) , *)
    (*     (forall n0 : nat, *)
    (*      m0 <= n0 -> *)
    (*      forall e' : Names.Exp F, *)
    (*      proj1_sig e' = proj1_sig (exist Universal_Property'_fold m m_UP') -> *)
    (*      exists Mn : ME {vn : Names.Value V | SubValueC V SV m' vn}, *)
    (*        fmap (@proj1_sig _ _) Mn = bevalM V F ME n0 e') ->  *)
    (*     (forall n0 : nat, *)
    (*      m0 <= n0 -> *)
    (*      forall e' : Names.Exp F, *)
    (*      proj1_sig e' = proj1_sig (exist Universal_Property'_fold n n_UP') -> *)
    (*      exists Mn : ME {vn : Names.Value V | SubValueC V SV n' vn}, *)
    (*        fmap (@proj1_sig _ _) Mn = bevalM V F ME n0 e') -> *)
    (*     let vm := (match isVI (proj1_sig m') with *)
    (*        | Some m'0 => *)
    (*          match isVI (proj1_sig n') with *)
    (*            | Some n'0 => return_ (vi' (m'0 + n'0)) *)
    (*            | None => *)
    (*              if isBot V (proj1_sig m') *)
    (*                then return_ (bot' V) *)
    (*                else *)
    (*                  if isBot V (proj1_sig n') *)
    (*                    then return_ (bot' V) *)
    (*                    else return_ (stuck' 4) *)
    (*          end *)
    (*        | None => *)
    (*          if isBot V (proj1_sig m') *)
    (*            then return_ (bot' V) *)
    (*            else *)
    (*              if isBot V (proj1_sig n') *)
    (*                then return_ (bot' V) *)
    (*                else return_ (stuck' 4) *)
    (*      end) in  *)
    (*     (forall n0 : nat, *)
    (*       S m0 <= n0 -> *)
    (*       forall e' : sig Universal_Property'_fold, *)
    (*         proj1_sig e' = in_t (inj (Add (Fix F) m n)) -> *)
    (*         exists Mn : ME {vn : Names.Value V | SubValueC V SV vm vn}, *)
    (*           fmap (@proj1_sig _ _) Mn = *)
    (*           boundedFix_UP n0 f_algebra (return_ (bot' V)) e')). *)
    (*   simpl; intros; destruct n0. *)
    (*   inversion H4. *)
    (*   assert (m0 <= n0) by auto with arith; clear H4. *)
    (*   simpl; rewrite H5; WF_FAlg_rewrite. *)
    (*   repeat erewrite bF_UP_in_out. *)
    (*   destruct (H2 _ H6 (exist _ _ m_UP') (refl_equal _)). *)
    (*   destruct (H3 _ H6 (exist _ _ n_UP') (refl_equal _)). *)
    (*   unfold bevalM, Exp, UP'_F in H4, H7.  *)
    (*   cut (forall (vm : {vm : Names.Value V | SubValueC V SV m' vm}) *)
    (*     (vn : {vn : Names.Value V | SubValueC V SV n' vn}), *)
    (*     SubValueC V SV *)
    (*     (match isVI (proj1_sig m') with *)
    (*       | Some m'0 => *)
    (*         match isVI (proj1_sig n') with *)
    (*           | Some n'0 => vi' (m'0 + n'0) *)
    (*           | None => *)
    (*             if isBot V (proj1_sig m') *)
    (*               then bot' V *)
    (*               else *)
    (*                 if isBot V (proj1_sig n') then bot' V else stuck' 4 *)
    (*         end *)
    (*       | None => *)
    (*         if isBot V (proj1_sig m') *)
    (*           then bot' V *)
    (*           else if isBot V (proj1_sig n') then bot' V else stuck' 4 *)
    (*     end) *)
    (*     (match isVI (proj1_sig (proj1_sig vm)) with *)
    (*        | Some m'0 => *)
    (*          match isVI (proj1_sig (proj1_sig vn)) with *)
    (*            | Some n'0 => vi' (m'0 + n'0) *)
    (*            | None => *)
    (*              if isBot V (proj1_sig (proj1_sig vm)) *)
    (*                then bot' V *)
    (*                else *)
    (*                  if isBot V (proj1_sig (proj1_sig vn)) then bot' V else stuck' 4 *)
    (*          end *)
    (*        | None => *)
    (*          if isBot V (proj1_sig (proj1_sig vm)) *)
    (*            then bot' V *)
    (*            else if isBot V (proj1_sig (proj1_sig vn)) then bot' V else stuck' 4 *)
    (*      end)). *)

    (*   intros; unfold evalMR, Exp, Names.Exp, UP'_F in *|-*; *)
    (*     rewrite <- H4, <- H7. *)
    (*   exists (do m' <- x1; *)
    (*           do n' <- x2; *)
    (*             return_ (exist _ _ (H8 m' n'))). *)
    (*   repeat (rewrite fmap_m; rewrite <- associativity). *)
    (*   apply f_equal; apply functional_extensionality; intros. *)
    (*   rewrite <- left_unit. *)
    (*   rewrite <- associativity. *)
    (*   repeat (rewrite fmap_m; rewrite <- associativity). *)
    (*   apply f_equal; apply functional_extensionality; intros. *)
    (*   repeat rewrite <- left_unit; simpl. *)
    (*   caseEq (isVI (proj1_sig (proj1_sig x3)) ). *)
    (*   caseEq (isVI (proj1_sig (proj1_sig x4))); auto. *)
    (*   caseEq (isBot V (proj1_sig (proj1_sig x3))); auto. *)
    (*   caseEq (isBot V (proj1_sig (proj1_sig x4))); auto. *)
    (*   caseEq (isBot V (proj1_sig (proj1_sig x3))); auto. *)
    (*   caseEq (isBot V (proj1_sig (proj1_sig x4))); auto. *)
    (*   Focus 2. *)
    (*   exists (do m' <- x;  *)
    (*           do n' <- x0; *)
    (*             (return_ (exist _ _ (H2 (proj1_sig m') (proj1_sig n') *)
    (*               (proj2_sig m') (proj2_sig n'))))). *)
    (*   repeat (rewrite fmap_m; rewrite <- associativity). *)
    (*   apply f_equal; apply functional_extensionality; intros. *)
    (*   rewrite <- left_unit. *)
    (*   repeat (rewrite fmap_m; rewrite <- associativity). *)
    (*   rewrite <- associativity. *)
    (*   apply f_equal; apply functional_extensionality; intros. *)
    (*   repeat rewrite <- left_unit; simpl. *)
    (*   destruct x1 as [[vm vm_UP'] P_vm]; *)
    (*     destruct x2 as [[vn vn_UP'] P_vn]; simpl. *)
    (*   caseEq (isVI vm). *)
    (*   caseEq (isVI vn); auto. *)
    (*   caseEq (isBot V vm); auto. *)
    (*   caseEq (isBot V vn); auto. *)
    (*   caseEq (isBot V vm); auto. *)
    (*   caseEq (isBot V vn); auto. *)
    (*   intros; destruct vm as [m'' Sub_m'_m''];  *)
    (*     destruct vn as [n'' Sub_n'_n'']; simpl.       *)
    (*   caseEq (project (G := NatValue) (proj1_sig m')). *)
    (*   unfold isVI at 1, Names.Exp; rewrite H8. *)
    (*   destruct n1. *)
    (*   generalize (inj_prj _ _ H8); rename H8 into H2'; intros. *)
    (*   assert (proj1_sig m'' = vi n1) as Eval_m. *)
    (*   unfold SubValueC in Sub_m'_m''; apply (SV_invertVI _ Sub_m'_m''); simpl. *)
    (*   unfold vi, vi', inject'; rewrite <- H8; rewrite in_out_UP'_inverse; eauto. *)
    (*   exact (proj2_sig _). *)
    (*   unfold isVI at 1. *)
    (*   caseEq (project (G := NatValue) (proj1_sig n')). *)
    (*   destruct n2. *)
    (*   rewrite Eval_m. *)
    (*   generalize (inj_prj _ _ H9); rename H9 into H3'; intros. *)
    (*   assert (proj1_sig n'' = vi n2) as Eval_n.  *)
    (*   unfold SubValueC in Sub_n'_n''; apply (SV_invertVI _ Sub_n'_n''); simpl.       *)
    (*   unfold vi, vi', inject'; rewrite <- H9; rewrite in_out_UP'_inverse; eauto. *)
    (*   exact (proj2_sig _). *)
    (*   unfold isVI; rewrite Eval_n. *)
    (*   unfold vi, vi', project; simpl; repeat rewrite out_in_fmap. *)
    (*   repeat rewrite wf_functor; simpl; repeat rewrite prj_inj. *)
    (*   apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; eauto. *)
    (*   caseEq (isBot V (proj1_sig m')). *)
    (*   apply inject_i; econstructor; eauto. *)
    (*   rewrite Eval_m; unfold isVI at 1, project, vi; simpl. *)
    (*   repeat rewrite out_in_fmap. *)
    (*   repeat rewrite wf_functor; simpl; repeat rewrite prj_inj. *)
    (*   caseEq (project (G := NatValue) (proj1_sig n'')). *)
    (*   unfold isVI; rewrite H11. *)
    (*   destruct n2. *)
    (*   apply project_inject in H11; auto with typeclass_instances. *)
    (*   destruct (SV_invertVI' _ Sub_n'_n'' _ H11) as [n_eq_vi | n_eq_bot]; *)
    (*     simpl in *|-.             *)
    (*   rewrite n_eq_vi in H9. *)
    (*   unfold vi, project in H9; simpl in H9; rewrite out_in_fmap in H9; *)
    (*     repeat rewrite wf_functor in H9; rewrite prj_inj in H9; *)
    (*       discriminate. *)
    (*   rewrite n_eq_bot; unfold isBot, bot, project; simpl; *)
    (*     rewrite out_in_fmap; repeat rewrite wf_functor; *)
    (*       rewrite prj_inj; simpl; *)
    (*         apply inject_i ; constructor; eauto. *)
    (*   exact (proj2_sig _). *)
    (*   unfold isVI; rewrite H11. *)
    (*   caseEq (project (G := BotValue) (vi n1)). *)
    (*   destruct b; generalize (inj_prj _ _ H12); intro.  *)
    (*   assert (vi n1 = bot _) by *)
    (*     (unfold vi, vi', bot, bot', inject' at -1; rewrite <- H13; *)
    (*       rewrite in_out_UP'_inverse; eauto with typeclass_instances). *)
    (*   unfold vi, bot in H14. *)
    (*   elimtype False; eapply (inject_discriminate Dis_VI_Bot _ _ H14). *)
    (*   unfold isBot; unfold vi, vi' in H12; simpl in H12; *)
    (*     rewrite wf_functor in H12; simpl in H12; rewrite H12. *)
    (*   caseEq (project (G := BotValue) (proj1_sig n'')). *)
    (*   destruct b. *)
    (*   apply project_inject in H13. *)
    (*   generalize (SV_invertBot _ SV _ _ Sub_n'_n'' H13); simpl; intro H8'; *)
    (*     rewrite H8'. *)
    (*   apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; simpl;  *)
    (*     eauto. *)
    (*   unfold bot, project, bot'; simpl;  *)
    (*     rewrite out_in_fmap; repeat rewrite wf_functor; *)
    (*       rewrite prj_inj; simpl; rewrite wf_functor; simpl; auto. *)
    (*   auto with typeclass_instances. *)
    (*   exact (proj2_sig _). *)
    (*   caseEq (project (G := BotValue) (proj1_sig n')). *)
    (*   destruct b. *)
    (*   apply inject_i; econstructor; eauto. *)
    (*   apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; simpl;  *)
    (*     eauto. *)
    (*   unfold isVI at 1; rewrite H8. *)
    (*   caseEq (project (G := BotValue) (proj1_sig m')). *)
    (*   unfold isBot at 1; rewrite H9. *)
    (*   destruct b. *)
    (*   apply inject_i; econstructor; eauto. *)
    (*   unfold isBot at 1; rewrite H9. *)
    (*   caseEq (project (G := BotValue) (proj1_sig n')). *)
    (*   unfold isBot at 1; rewrite H10. *)
    (*   destruct b. *)
    (*   apply inject_i; econstructor; eauto. *)
    (*   unfold isBot at 1; rewrite H10. *)
    (*   caseEq (project (G := NatValue) (proj1_sig m'')); *)
    (*   unfold isVI; rewrite H11. *)
    (*   destruct n1. *)
    (*   apply project_inject in H11. *)
    (*   destruct (SV_invertVI' _ Sub_m'_m'' _ H11) as [m_eq_vi | m_eq_bot]; *)
    (*     simpl in *|-.            *)
    (*   rewrite m_eq_vi in H8. *)
    (*   rewrite project_vi_vi in H8; discriminate. *)
    (*   rewrite m_eq_bot in H9. *)
    (*   unfold project, bot in H9; simpl in H9. *)
    (*   rewrite out_in_fmap in H9; repeat rewrite wf_functor in H9; *)
    (*       rewrite prj_inj in H9; simpl in H9; discriminate.  *)
    (*   auto with typeclass_instances. *)
    (*   exact (proj2_sig _). *)
    (*   caseEq (project (G := BotValue) (proj1_sig m'')). *)
    (*   unfold isBot at 1; rewrite H12. *)
    (*   destruct b; apply project_inject in H12. *)
    (*   generalize (SV_invertBot _ SV _ _ Sub_m'_m'' H12); simpl; intro H8'; *)
    (*     rewrite H8' in H9. *)
    (*   unfold project, bot in H9; simpl in H9. *)
    (*   rewrite out_in_fmap in H9; repeat rewrite wf_functor in H9; *)
    (*       rewrite prj_inj in H9; simpl in H9; discriminate.  *)
    (*   auto with typeclass_instances. *)
    (*   exact (proj2_sig _). *)
    (*   unfold isBot; rewrite H12. *)
    (*   caseEq (project (G := BotValue) (proj1_sig n'')). *)
    (*   destruct b; apply project_inject in H13. *)
    (*   generalize (SV_invertBot _ SV _ _ Sub_n'_n'' H13); simpl; intro H8'; *)
    (*     rewrite H8' in H10. *)
    (*   unfold project, bot in H10; simpl in H10. *)
    (*   rewrite out_in_fmap in H10; repeat rewrite wf_functor in H10; *)
    (*       rewrite prj_inj in H10; simpl in H10; discriminate.  *)
    (*   auto with typeclass_instances. *)
    (*   exact (proj2_sig _). *)
    (*   apply (inject_i (subGF := Sub_SV_refl_SV)); constructor; simpl;  *)
    (*     eauto. *)
    (* Qed. *)


    (* Lemma WF_ind_alg_Arith (Name : Set)
    (P : forall e : Fix F, Universal_Property'_fold e -> Prop)
    (H : forall n, UP'_P P (lit n))
    (H0 : forall m n
      (IHm : UP'_P P m)
      (IHn : UP'_P P n),
      UP'_P P (@add m n (proj1_sig IHm) (proj1_sig IHn)))
    {Sub_Arith_F' : Arith :<: F} :
    (forall a, inj (Sub_Functor := Sub_Arith_F) a =
      inj (A := (Fix F)) (Sub_Functor := Sub_Arith_F') a) ->
      WF_Ind (Name := Name) {| p_algebra := ind_alg_Arith P H H0|}.
      constructor; intros.
      simpl; unfold ind_alg_Arith; destruct e; simpl.
      unfold lit; simpl; rewrite wf_functor; simpl; apply f_equal; eauto.
      unfold add; simpl; rewrite wf_functor; simpl; apply f_equal; eauto.
    Defined. *)

    (* Context {eval_continuous_Exp_E :
      PAlgebra EC_ExpName F F (UP'_P (eval_continuous_Exp_P V F SV))}.

  (* ============================================== *)
  (* WELL-FORMED NAT VALUES                         *)
  (* ============================================== *)

    Variable WFV : (WFValue_i D V -> Prop) -> WFValue_i D V -> Prop.
    Variable funWFV : iFunctor WFV.

    (** Natrual Numbers are well-formed **)

    Inductive WFValue_VI (WFV : WFValue_i D V -> Prop) : WFValue_i D V -> Prop :=
    | WFV_VI : forall n v T,
      proj1_sig v = vi n ->
      proj1_sig T = tnat ->
      WFValue_VI WFV (mk_WFValue_i D V v T).

    Definition ind_alg_WFV_VI (P : WFValue_i D V -> Prop)
      (H : forall n v T veq Teq, P (mk_WFValue_i _ _ v T))
      i (e : WFValue_VI P i) : P i :=
      match e in WFValue_VI _ i return P i with
        | WFV_VI n v T veq Teq => H n v T veq Teq
      end.

    Definition WFV_VI_ifmap (A B : WFValue_i D V -> Prop) i (f : forall i, A i -> B i)
      (WFV_a : WFValue_VI A i) : WFValue_VI B i :=
      match WFV_a in (WFValue_VI _ s) return (WFValue_VI B s)
        with
        | WFV_VI n v T veq Teq => WFV_VI B n v T veq Teq
      end.

    Global Instance iFun_WFV_VI : iFunctor WFValue_VI.
      constructor 1 with (ifmap := WFV_VI_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable Sub_WFV_VI_WFV : Sub_iFunctor WFValue_VI WFV.

     Global Instance WFV_proj1_a_VI :
       iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V WFV) WFValue_VI.
       econstructor; intros.
       unfold iAlgebra; intros; unfold WFV_proj1_a_P.
       inversion H; subst; simpl; intros.
       apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; simpl; eauto.
       rewrite H3; eauto.
     Defined.

     Global Instance WFV_proj1_b_VI :
       iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V WFV) WFValue_VI.
       econstructor; intros.
       unfold iAlgebra; intros; unfold WFV_proj1_b_P.
       inversion H; subst; simpl; intros.
       apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; simpl; eauto.
       rewrite H3; eauto.
     Defined.

    (* Inversion principles for Well-formed natural numbers. *)
    Definition WF_invertVI_P (i : WFValue_i D V) :=
      proj1_sig (wfv_b _ _ i) = tnat ->
      WFValue_VI (iFix WFV) i \/ (proj1_sig (wfv_a D V i) = bot V).

    Inductive WF_invertVI_Name := wfv_invertvi_name.
    Context {WF_invertVI_WFV :
      iPAlgebra WF_invertVI_Name WF_invertVI_P WFV}.

    Global Instance WF_invertVI_VI :
      iPAlgebra WF_invertVI_Name WF_invertVI_P WFValue_VI.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertVI_P.
      inversion H; subst; simpl; intros.
      left; econstructor; eassumption.
    Defined.

    Global Instance WF_invertVI_Bot :
      iPAlgebra WF_invertVI_Name WF_invertVI_P (WFValue_Bot _ _).
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertVI_P.
      inversion H; subst; simpl; intros.
      inversion H; subst; rewrite H3; right; reflexivity.
    Defined.

    Definition WF_invertVI := ifold_ WFV _ (ip_algebra (iPAlgebra := WF_invertVI_WFV)).

    Context {WFV_proj1_a_WFV :
      iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V WFV) WFV}.
    Context {WFV_proj1_b_WFV :
       iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V WFV) WFV}.

    Lemma Arith_eval_Soundness_H
      (typeof_R eval_R : Set) typeof_rec eval_rec
      {eval_F' : FAlgebra EvalName eval_R (evalR V) F}
      {WF_eval_F' : @WF_FAlgebra EvalName _ _ Arith F
        Sub_Arith_F (MAlgebra_eval_Arith _) (eval_F')} :
      forall n : nat,
        forall gamma'' : Env (Names.Value V),
          forall T : Names.DType D,
            Arith_typeof typeof_R typeof_rec (Lit _ n) = Some T ->
            WFValueC D V WFV (Arith_eval eval_R eval_rec (Lit _ n) gamma'') T.
      intros n gamma'' T H4; intros.
      apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; eauto.
      simpl.
      unfold vi, vi', inject; simpl; eauto.
      unfold typeof, mfold, lit in H4; simpl in H4.
      injection H4; intros; subst.
      reflexivity.
    Defined.

   Lemma Arith_eval_Soundness_H0
     (typeof_R eval_R : Set) typeof_rec eval_rec
     {eval_F' : FAlgebra EvalName eval_R (evalR V) F}
     {WF_eval_F' : @WF_FAlgebra EvalName _ _ Arith F
        Sub_Arith_F (MAlgebra_eval_Arith _) (eval_F')} :
      forall (a b : typeof_R) (a' b' : eval_R),
        forall gamma'' : Env (Names.Value V),
          (forall T : Names.DType D,
            typeof_rec a = Some T ->
            WFValueC D V WFV (eval_rec a' gamma'') T) ->
          (forall T : Names.DType D,
            typeof_rec b = Some T ->
            WFValueC D V WFV (eval_rec b' gamma'') T) ->
          forall T : Names.DType D,
            Arith_typeof typeof_R typeof_rec (Add _ a b) = Some T ->
            WFValueC D V WFV (Arith_eval eval_R eval_rec (Add _ a' b') gamma'') T.
     simpl; intros a b a' b' gamma'' IH_a IH_b T H4.
      caseEq (typeof_rec a); intros; rename H into typeof_a;
        unfold typeofR in typeof_a, H4; rewrite typeof_a in H4;
          try discriminate.
      caseEq (typeof_rec b); intros; rename H into typeof_b;
        unfold typeofR in typeof_b, H4; rewrite typeof_b in H4;
          try discriminate.
      caseEq (isTNat (proj1_sig d)); intros; rename H into d_eq; rewrite
        d_eq in H4; try discriminate.
      caseEq (isTNat (proj1_sig d0)); intros; rename H into d0_eq; rewrite
        d0_eq in H4; try discriminate.
      injection H4; intros; subst; clear H4.
      unfold isTNat in d_eq, d0_eq.
      caseEq (project (proj1_sig d)); intros; rewrite H in d_eq;
        try discriminate; clear d_eq; rename H into d_eq; destruct a0.
      caseEq (project (proj1_sig d0)); intros; rewrite H in d0_eq;
        try discriminate; clear d0_eq; rename H into d0_eq; destruct a0.
      apply project_inject in d_eq; apply project_inject in d0_eq;
        eauto with typeclass_instances.
      generalize (IH_a _ typeof_a) as WF_a;
      generalize (IH_b _ typeof_b) as WF_b; intros.
      unfold WFValueC in *|-*.
      destruct (WF_invertVI _ WF_a d_eq) as [beval_a' | beval_a'];
        inversion beval_a'; subst.
      rewrite H1; unfold isVI, project, vi, vi', inject'; simpl;
        rewrite out_in_fmap; repeat rewrite wf_functor; simpl; rewrite prj_inj.
      destruct (WF_invertVI _ WF_b d0_eq) as [beval_b' | beval_b'];
        inversion beval_b'; subst.
      rewrite H3; unfold isVI, project, vi, vi', inject'; simpl;
        rewrite out_in_fmap; repeat rewrite wf_functor; simpl; rewrite prj_inj.
      apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; eauto.
      simpl; rewrite wf_functor; simpl; eauto.
      unfold vi, vi', inject'; simpl; eauto;
        rewrite wf_functor; simpl; reflexivity.
      rewrite H0; unfold bot, isVI, project, inject, inject'; simpl;
        rewrite out_in_fmap; repeat rewrite wf_functor; simpl; unfold Bot_fmap.
      caseEq (prj (Sub_Functor := Sub_NatValue_V) (A:= (sig (@Universal_Property'_fold V _)))
        (inj (Bot (sig Universal_Property'_fold)))).
      elimtype False; eapply (inject_discriminate Dis_VI_Bot n0).
      unfold inject; simpl; apply inj_prj in H; erewrite <- H; reflexivity.
      caseEq (isBot V (in_t (inj (VI (Fix V) n)))).
      simpl in beval_b'.
      apply sym_eq in H0; apply sym_eq in d0_eq.
      generalize (WFV_proj1_a D V WFV _ _ WF_b _ _ H0).
      simpl; intros WF_b'; generalize (WFV_proj1_b D V WFV _ _ WF_b' _ _ d0_eq); simpl.
      assert (exist Universal_Property'_fold (bot V) (UP'_bot V) = bot' V) by
        (unfold bot, UP'_bot; destruct bot'; reflexivity).
      assert (exist Universal_Property'_fold tnat UP'_tnat = tnat') by
        (unfold tnat, UP'_tnat; destruct tnat'; reflexivity).
      unfold tnat, tnat', inject' at 1 in H5.
      unfold inject.
      rewrite H5, H4; auto.
      unfold isBot, project; simpl; rewrite out_in_fmap;
        repeat rewrite wf_functor; simpl; unfold Bot_fmap;
          rewrite prj_inj.
      simpl in beval_b'.
      apply sym_eq in H0; apply sym_eq in d0_eq.
      unfold eval in WF_b; generalize (WFV_proj1_a D V WFV _ _ WF_b _ _ H0).
      simpl; intros WF_b'; generalize (WFV_proj1_b D V WFV _ _ WF_b' _ _ d0_eq); simpl.
      assert (exist Universal_Property'_fold (bot V) (UP'_bot V) = bot' V) by
        (unfold bot, UP'_bot; destruct bot'; reflexivity).
      assert (exist Universal_Property'_fold tnat UP'_tnat = tnat') by
        (unfold tnat, UP'_tnat; destruct tnat'; reflexivity).
      unfold tnat, tnat', inject' at 1 in H5.
      unfold inject.
      rewrite H5, H4; auto.
      caseEq (isVI (proj1_sig (eval_rec a' gamma''))).
      unfold isVI in H; rewrite H0 in H.
      unfold bot, isVI, project, inject, inject' in H; simpl in H;
        rewrite out_in_fmap in H; repeat rewrite wf_functor in H; simpl in H;
          unfold Bot_fmap in H.
      caseEq (prj (Sub_Functor := Sub_NatValue_V) (A:= (sig (@Universal_Property'_fold V _)))
        (inj (Bot (sig Universal_Property'_fold)))).
      elimtype False; eapply (inject_discriminate Dis_VI_Bot n0);
        apply inj_prj in H1; unfold inject; rewrite <- H1; reflexivity.
      rewrite H1 in H; discriminate.
      rewrite H0.
      unfold bot, isBot, project, inject; simpl; rewrite out_in_fmap;
        repeat rewrite wf_functor; simpl; unfold Bot_fmap; simpl.
      simpl in beval_a'.
      apply sym_eq in H0; apply sym_eq in d_eq.
      generalize (WFV_proj1_a D V WFV _ _ WF_a _ _ H0).
      simpl; intros WF_a'; generalize (WFV_proj1_b D V WFV _ _ WF_a' _ _ d_eq); simpl.
      assert (exist Universal_Property'_fold (bot V) (UP'_bot V) = bot' V) by
        (unfold bot, UP'_bot; destruct bot'; reflexivity).
      assert (exist Universal_Property'_fold tnat UP'_tnat = tnat') by
        (unfold tnat, UP'_tnat; destruct tnat'; reflexivity).
      unfold tnat, tnat', inject' at 1 in H2.
      unfold inject.
      rewrite H2, H1, prj_inj; auto.
      exact (proj2_sig d0).
      exact (proj2_sig d0).
      exact (proj2_sig d).
      exact (proj2_sig d0).
    Defined.

    Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR _) F}.
    Context {WF_typeof_F : forall T, @WF_FAlgebra TypeofName T _ _ _
      Sub_Arith_F (MAlgebra_typeof_Arith T) (Typeof_F _)}.
    Context {WF_Value_continous_alg :
      iPAlgebra WFV_ContinuousName (WF_Value_continuous_P D V WFV) SV}.

    Global Instance Arith_eval_Soundness_alg
      (P_bind : Set)
      (P : P_bind -> Env Value -> Prop)
      (E' : Set -> Set)
      {Fun_E' : Functor E'}
      {Sub_Arith_E' : Arith :<: E'}
      {WF_Fun_E' : WF_Functor _ _ Sub_Arith_E' _ _}
      {Typeof_E' : forall T, FAlgebra TypeofName T (typeofR _) E'}
      {WF_typeof_E' : forall T, @WF_FAlgebra TypeofName T _ _ _
        Sub_Arith_E' (MAlgebra_typeof_Arith T) (Typeof_E' _)}
      (pb : P_bind)
      (eval_rec : Exp -> evalR V)
      (typeof_rec : UP'_F E' -> typeofR D)
      :
      P2Algebra eval_Soundness_alg_Name Arith E' F
      (UP'_P2 (@eval_alg_Soundness_P D _ V _ _ _ WFV _ P
        _ _ pb typeof_rec eval_rec (f_algebra (FAlgebra := Typeof_E' _ ))
        (f_algebra (FAlgebra := eval_F)))).
    Proof.
      apply ind_P2Alg_Arith; try eassumption.
      unfold eval_alg_Soundness_P, UP'_P2; intros.
      constructor.
      exact (conj (proj2_sig (inject' (Lit _ n))) (proj2_sig (lit' n))).
      unfold lit, lit', inject; simpl.
      repeat rewrite out_in_fmap.
      repeat rewrite wf_functor.
      intros gamma'' WF_gamma'' IHa.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold Arith_fmap.
      intros.
      eapply Arith_eval_Soundness_H.
      apply WF_eval_F.
      apply H.
      (* Add Case *)
      unfold eval_alg_Soundness_P, UP'_P2; intros.
      destruct m as [m1 m2]; destruct n as [n1 n2];
        destruct IHm as [[UP'_m1 UP'_m2] IHm];
          destruct IHn as [[UP'_n1 UP'_n2] IHn]; simpl in *|-*.
      constructor.
      split; unfold inject; exact (proj2_sig _).
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj gamma'' WF_gamma'' IHa.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold Arith_fmap.
      intros T; eapply Arith_eval_Soundness_H0.
      apply WF_eval_F.
      apply (IHa _ _ WF_gamma'' (_, exist _ _ UP'_m2)); simpl;
        intros; apply (IHm eval_rec_proj typeof_rec_proj _ WF_gamma'' IHa);
          intros; auto; rewrite <- (in_out_UP'_inverse _ _ m1); auto.
      apply (IHa _ _ WF_gamma'' (_, exist _ _ UP'_n2)); simpl;
        intros; apply (IHn eval_rec_proj typeof_rec_proj _ WF_gamma'' IHa);
          intros; auto; rewrite <- (in_out_UP'_inverse _ _ n1); auto.
    Defined. *)

End Arith.

 (* Hint Extern 1 (iPAlgebra SV_invertVI_Name (SV_invertVI_P _) _) =>
      constructor; unfold iAlgebra; unfold SV_invertVI_P; intros i H n H0;
        inversion H; subst; simpl in H0; revert H0;
          match goal with H : proj1_sig ?v = _ |- proj1_sig ?v = _ -> _ => rewrite H end; intros;
            elimtype False; apply (inject_discriminate _ _ _ H0).

  Hint Extern 1 (iPAlgebra SV_invertVI'_Name (SV_invertVI'_P _) _) =>
      constructor; unfold iAlgebra; unfold SV_invertVI'_P; intros i H n H0;
        inversion H; subst; simpl in H0; revert H0;
          match goal with H : proj1_sig ?v = _ |- proj1_sig ?v = _ -> _ => rewrite H end; intros;
            elimtype False; apply (inject_discriminate _ _ _ H0). *)

  Hint Extern 5 (iPAlgebra WF_invertVI_Name (WF_invertVI_P _ _ _ _) _) =>
    constructor; unfold iAlgebra; let H := fresh in
      intros i H; unfold WF_invertVI_P;
        inversion H; simpl;
          match goal with
            eq_H0 : proj1_sig ?T = _ |- proj1_sig ?T = _ -> _ =>
              intro eq_H; rewrite eq_H in eq_H0;
                elimtype False; first [apply (inject_discriminate _ _ _ eq_H0) |
                  apply sym_eq in eq_H0; apply (inject_discriminate _ _ _ eq_H0)]
          end : typeclass_instances.

(* Hint Extern 0 =>
  intros; match goal with
            |- (PAlgebra eval_Soundness_alg_Name
              (sig (UP'_P2 (eval_alg_Soundness_P _ _ _ _ _ _ _ _ _ _ _ _ _ _))) Arith) =>
            eapply Arith_eval_Soundness_alg; eauto with typeclass_instances
          end : typeclass_instances. *)


(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
