Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.
Require Import Bool.
Require Import Exception.

Section Exception_Bool.

  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.
  Context {Sub_BType_D : BType :<: D}.

  Variable F : Set -> Set.
  Context {Fun_F : Functor F}.
  Context {Sub_ExceptE_F : ExceptE D :<: F}.
  Context {Sub_Bool_F : Bool :<: F}.

  (* Constructor + Universal Property. *)
   Context {WF_Sub_ExceptE_F : WF_Functor _ _ Sub_ExceptE_F _ _}.
   Context {WF_Sub_Bool_F : WF_Functor _ _ Sub_Bool_F _ _}.

   Variable V : Set -> Set.
   Context {Fun_V : Functor V}.
   Context {Sub_StuckValue_V : StuckValue :<: V}.
   Context {Sub_NatValue_V : BoolValue :<: V}.

  Variable MT : Set -> Set.
  Context `{Fail_MT : FailMonad MT}.

  Context {Inj_MT : InjMonad MT}.
  Context {Reasonable_MT : Reasonable_Monad MT}.
  Context {MT_eq_dec : forall (A : Set) (mta : MT A),
    {exists a, mta = return_ a} + {mta = fail}}.

  Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.
  Context {eq_DType_eq_DT : PAlgebra eq_DType_eqName D D (UP'_P (eq_DType_eq_P D))}.
  Context {eq_DType_neq_D : PAlgebra eq_DType_neqName D D (UP'_P (eq_DType_neq_P D))}.

  Variable ME : Set -> Set.
  Context `{Exception_ME : Exception ME unit}.
  Context {Fail_ME : FailMonad ME}.

  Context {ME_eq_dec : forall (A : Set) (mte : ME A),
    (exists a, mte = return_ a) \/ (mte = fail) \/ (mte = throw tt)}.
  Context {Reasonable_ME : Reasonable_Monad ME}.
  Context {fail_neq_throw : forall A, fail (M := ME) <> throw (A := A) tt}.

  Context {evalM_F : FAlgebra EvalName (Exp F) (evalMR V ME) F}.
  Context {WF_eval_ExceptE_F : @WF_FAlgebra EvalName _ _ (ExceptE D) F
    Sub_ExceptE_F (MAlgebra_evalM_ExceptE D V ME _) evalM_F}.
  Context {WF_eval_Bool_F : @WF_FAlgebra EvalName _ _ Bool F
    Sub_Bool_F (MAlgebra_evalM_Bool V ME _) evalM_F}.

  Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) F}.
  Context {WF_ExceptE_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ (ExceptE D) F
    Sub_ExceptE_F (MAlgebra_typeof_ExceptE _ MT _) (Typeof_F T)}.
  Context {WF_Bool_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ Bool F
    Sub_Bool_F (MAlgebra_typeof_Bool _ MT _) (Typeof_F T)}.

    Context {evalM_F' : forall T, FAlgebra EvalName T (evalMR V ME) F}.
    Context {WF_ExceptE_eval_F' : forall T, @WF_FAlgebra EvalName _ _ (ExceptE D) F
      Sub_ExceptE_F (MAlgebra_evalM_ExceptE D V ME _) (evalM_F' T)}.
    Context {WF_Bool_eval_F' : forall T, @WF_FAlgebra EvalName _ _ Bool F
      Sub_Bool_F (MAlgebra_evalM_Bool V ME _) (evalM_F' T)}.

    Variable TypContext : Set.
    
    Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
    Context {funWFV : iFunctor WFV}.

    Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
    Context {funWFVM : iFunctor WFVM}.
    
    Context {TypContextCE : ConsExtensionC TypContext}.

    Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME _ WFV) WFVM}.
    Context {Sub_WFVM_Except_WFVM : Sub_iFunctor (WFValueM_Except D V MT ME _) WFVM}.
    
    Global Instance WFVM_Except_eval_Soundness_Ifi :
      iPAlgebra Bool_eval_Soundness_Ifi_Name
      (Bool_eval_Soundness_Ifi_P D MT V ME TypContext WFVM) (WFValueM_Except D V MT ME TypContext).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
        try eassumption; unfold Bool_eval_Soundness_Ifi_P; split; simpl; intros.
    (* WFVM_Throw' *)
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 1.
      rewrite bind_throw; apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
        econstructor.
    (* WFVM_Catch' *)
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2;
        eauto.
      apply (proj1 H0).
      intros t Sigma' ConsExt_Sig_Sig'; apply (proj1 (H1 t Sigma' ConsExt_Sig_Sig')).
      repeat rewrite <- associativity.
      rewrite catch_msimp.
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
      intros; apply f_equal.
      repeat rewrite kt_eq_msimp with
        (f := fun c => if isTBool D c
          then
            do Tt <- mtT;
              do Te <- meT;
                (if eq_DType D (proj1_sig Tt) Te then return_ Tt else fail)
          else fail).
      erewrite kt_eq; eauto.
      rewrite associativity, <- catch_msimp, associativity; apply H0; auto.
      intros; rewrite associativity, <- catch_msimp, associativity; apply H1; eauto.
      intros; eapply H2; eapply ConsExtension_trans; eauto.
      intros; eapply H3; eapply ConsExtension_trans; eauto.
    Qed.

  Global Instance WFVM_State_eval_Soundness_Ift :
    iPAlgebra Bool_eval_Soundness_Ift_Name
    (Bool_eval_Soundness_Ift_P D MT V ME TypContext WFVM) (WFValueM_Except D V MT ME TypContext).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try eassumption; unfold Bool_eval_Soundness_Ift_P; simpl; split; intros.
    (* WFVM_Throw *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 1.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
      econstructor; intros; eapply H0; eauto.
    (* WFVM_Put *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2;
      eauto.
    apply (proj1 H0).
    intros t Sigma' ConsExt_Sig_Sig'; apply (proj1 (H1 t Sigma' ConsExt_Sig_Sig')).
    repeat rewrite <- associativity.
    rewrite curry_msimp.
    rewrite catch_msimp.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
    intros.
    assert (forall T, fmap (@proj1_sig _ _) (
     do U <- do a <- mti; do b <- kT'; return_ (a, b);
      do c <- kT (snd U) T;
      do Te <- meT;
      (if eq_DType D (proj1_sig c) Te then return_ c else fail)) =
     do U <- do a <- mti; do b <- kT'; return_ (a, b);
       do c <- kT (snd U) T;
      do Te <- meT;
      (if eq_DType D (proj1_sig c) Te then return_ (proj1_sig c) else fail)) as mt_eq by
    (intros T''; rewrite fmap_m;
      rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
        rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
          rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
            destruct (eq_DType D (proj1_sig x0) x1);
              [rewrite <- left_unit; auto | rewrite bind_fail; auto]).
    repeat rewrite mt_eq.
    repeat rewrite <- (curry_msimp _ mti (fun _ => kT')
      (fun _ U => do c <- kT U _;
        do Te <- meT;
          (if eq_DType D (proj1_sig c) Te then return_ _ else fail))).
    apply f_equal; apply functional_extensionality; intros.
    generalize (kt_eq_msimp _ kT' (fun b => kT b T)
      (fun c : Fix D => do Te <- meT;
        (if eq_DType D c Te then return_ c else fail)) (@proj1_sig _ _)).
    generalize (kt_eq_msimp _ kT' (fun b => kT b T')
      (fun c : Fix D => do Te <- meT;
        (if eq_DType D c Te then return_ c else fail)) (@proj1_sig _ _)).
    unfold DType, Names.DType, UP'_F, DType;
    intros mt_eq' mt'_eq'; rewrite mt'_eq', mt_eq'.
    erewrite kt_eq; eauto.
    rewrite <- (curry_msimp _ mti (fun _ => kT')
      (fun _ U => do b <- mte;
        do c <- kT U b;
          do Te <- meT;
            (if eq_DType D (proj1_sig c) Te then return_ c else fail))).
    rewrite rewrite_do with (n := mti) (m' := (fun _ : C0 =>
           do Tt <- do U <- kT'; mte >>= kT U;
           do Te <- meT;
           (if eq_DType D (proj1_sig Tt) Te then return_ Tt else fail))).
    eapply H0; eauto.
    apply functional_extensionality; intros; rewrite <- associativity;
      apply f_equal; apply functional_extensionality; intros;
        rewrite <- associativity; auto.
    intros; rewrite <- (curry_msimp _ mti (fun _ => kT')
      (fun _ U => do b <- mth;
        do c <- kT U b;
          do Te <- meT;
            (if eq_DType D (proj1_sig c) Te then return_ c else fail))).
    rewrite rewrite_do with (n := mti) (m' := (fun _ : C0 =>
           do Tt <- do U <- kT'; mth >>= kT U;
           do Te <- meT;
           (if eq_DType D (proj1_sig Tt) Te then return_ Tt else fail))).
    eapply H1; eauto.
    intros; eapply H2; eapply ConsExtension_trans; eauto.
    apply functional_extensionality; intros; rewrite <- associativity;
      apply f_equal; apply functional_extensionality; intros;
        rewrite <- associativity; eauto.
  Qed.

  Global Instance WFVM_Except_eval_Soundness_Ife :
    iPAlgebra Bool_eval_Soundness_Ife_Name
    (Bool_eval_Soundness_Ife_P D MT V ME _ WFVM) (WFValueM_Except D V MT ME _).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try eassumption; unfold Bool_eval_Soundness_Ife_P; split; simpl; intros.
    (* WFVM_Throw *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
      econstructor; eauto.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
      econstructor; intros; eapply H0; eauto.
    (* WFVM_Put *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2;
      eauto.
    apply (proj1 H0).
    intros t Sigma' ConsExt_Sig_Sig'; apply (proj1 (H1 t Sigma' ConsExt_Sig_Sig')).
    rewrite catch_msimp.
    rewrite curry3_msimp.
    rewrite catch_msimp.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
    assert (forall T, fmap (@proj1_sig _ _)
      (do U <-
        do a <- mti; do b <- mtT; do c <- kT'; return_ (a, (b, c));
          do c <- kT (snd (snd U)) T;
            (if eq_DType D (proj1_sig (fst (snd U))) c
              then return_ (fst (snd U))
              else fail)) =
            do U <-
        do a <- mti; do b <- mtT; do c <- kT'; return_ (a, (b, c));
          do c <- kT (snd (snd U)) T;
            (if eq_DType D (proj1_sig (fst (snd U))) c
              then return_ (proj1_sig (fst (snd U)))
              else fail)) as mt_eq by
    (intros T''; rewrite fmap_m;
      rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
        rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
          destruct (eq_DType D (proj1_sig (fst (snd x))) x0);
            [rewrite <- left_unit; auto | rewrite bind_fail; auto]).
    intros; repeat rewrite mt_eq.
    intros; repeat rewrite <- curry3_msimp with
      (mta := mti) (mtb := fun _ => mtT)
      (mtc := fun _ _ => kT')
      (f := fun _ U U' => do c <- kT U' _;
      (if eq_DType D (proj1_sig U) c then return_ (proj1_sig U) else fail)).
    apply f_equal; apply functional_extensionality; intros.
    apply f_equal; apply functional_extensionality; intros.
    repeat rewrite associativity.
    rewrite rewrite_do with (M := MT) (n :=  do x1 <- kT'; kT x1 _) (m' := (fun c : Names.DType D =>
      (if eq_DType D (proj1_sig c) x0 then return_ (proj1_sig x0) else fail))).
    rewrite rewrite_do with (M := MT) (n :=  do x1 <- kT'; kT x1 T') (m' := (fun c : Names.DType D =>
      (if eq_DType D (proj1_sig c) x0 then return_ (proj1_sig x0) else fail))).
    repeat rewrite <- associativity,
      (kt_eq_msimp _ kT' (fun c => kT c _)
        (fun c0 => if eq_DType D c0 x0 then return_ (proj1_sig x0) else fail) (@proj1_sig _ _)).
    unfold DType, Names.DType, UP'_F, DType;
      erewrite kt_eq; eauto.
    apply functional_extensionality; intros;
      caseEq (eq_DType D (proj1_sig x0) x1); caseEq (eq_DType D (proj1_sig x1) x0); auto; elimtype False;
        [apply (eq_DType_neq _ _ _ H5); generalize (eq_DType_eq _ _ _ H4) |
          apply (eq_DType_neq _ _ _ H4); generalize (eq_DType_eq _ _ _ H5)];
        congruence.
    apply functional_extensionality; intros;
      caseEq (eq_DType D (proj1_sig x0) x1); caseEq (eq_DType D (proj1_sig x1) x0); auto; elimtype False;
        [apply (eq_DType_neq _ _ _ H5); generalize (eq_DType_eq _ _ _ H4) |
          apply (eq_DType_neq _ _ _ H4); generalize (eq_DType_eq _ _ _ H5)];
        congruence.
    rewrite <- curry3_msimp with (mta := mti) (mtb := fun _ => mtT)
      (mtc := fun _ _ => kT')
      (f := fun _ U U' => do b <- mte; do c <- kT U' b;
      (if eq_DType D (proj1_sig U) c then return_ U else fail)).
    rewrite associativity.
    rewrite rewrite_do with (n := mti >>= (fun _ : C0 => mtT))
      (m' := fun Tt => do Te <- do U <- kT'; mte >>= kT U;
        (if eq_DType D (proj1_sig Tt) Te then return_ Tt else fail)).
    rewrite <- associativity; eapply H0; auto.
    apply functional_extensionality; intros; rewrite <- associativity;
      apply f_equal; apply functional_extensionality; intros; rewrite <- associativity;
        apply f_equal; apply functional_extensionality; intros; apply f_equal;
          apply functional_extensionality; intros; auto.
    intros.
    rewrite <- curry3_msimp with (mta := mti) (mtb := fun _ => mtT)
      (mtc := fun _ _ => kT')
      (f := fun _ U U' => do b <- mth; do c <- kT U' b;
      (if eq_DType D (proj1_sig U) c then return_ U else fail)).
    rewrite associativity.
    rewrite rewrite_do with (n := mti >>= (fun _ : C0 => mtT))
      (m' := fun Tt => do Te <- do U <- kT'; mth >>= kT U;
        (if eq_DType D (proj1_sig Tt) Te then return_ Tt else fail)).
    rewrite <- associativity; eapply H1; eauto.
    intros; eapply H2; eapply ConsExtension_trans; eauto.
    apply functional_extensionality; intros; rewrite <- associativity;
      apply f_equal; apply functional_extensionality; intros; rewrite <- associativity;
        apply f_equal; apply functional_extensionality; intros; apply f_equal;
          apply functional_extensionality; intros; auto.
  Qed.

    (* Doesn't work...
       Global Instance WFVM_Except_eval_Soundness_Addn' :
    iPAlgebra Arith_eval_Soundness_Addn'_Name
    (Arith_eval_Soundness_Addn'_P V ME D MT WFV WFVM) (WFValueM_Except D V MT ME).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try assumption; unfold Arith_eval_Soundness_Addn'_P; simpl; intros.
    (* throw case *)
    rewrite bind_throw.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    (* catch case *)
    repeat rewrite <- associativity.
    assert (do tm <- fmap (@proj1_sig _ _) mt';
      do tn <- do x <- mte; (do th <- mth;
        (if eq_DType D (proj1_sig x) th then return_ x else fail)) >>= kT;
      (if isTNat D (proj1_sig tm)
        then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
        else fail) =
      (do tm <- fmap (@proj1_sig _ _) mt';
        do tn <- mte;
          (do th <- mth; (if eq_DType D (proj1_sig tn) th then return_ tn else fail)) >>=
          fun x => do tn <- kT x;
            (if isTNat D (proj1_sig x)
              then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
              else fail))).
    rewrite (eq_bind_fmap _ _ _ _ mt' _
      (fun tm => do tn <- typeof (Typeof_E := Typeof_F) D F MT n;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    apply f_equal; apply functional_extensionality; intros;
      repeat (rewrite <- associativity; apply f_equal;
        apply functional_extensionality; intros); auto.
    apply f_equal; apply functional_extensionality; intros.
    apply f_equal.

 as kT_eq by
    (apply f_equal; apply functional_extensionality; intros;
      (repeat (rewrite <- associativity; apply f_equal;
        apply functional_extensionality; intros); auto)); rewrite kT_eq; clear kT_eq.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
    intros; apply f_equal.
    rewrite (eq_bind_fmap _ _ _ _ (kT T) _
      (fun tm => do tn <- typeof (Typeof_E := Typeof_F) D F MT n;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    rewrite (eq_bind_fmap _ _ _ _ (kT T') _
      (fun tm => do tn <- typeof (Typeof_E := Typeof_F) D F MT n;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    unfold Names.DType, UP'_F; rewrite (kt_eq _ _ H3); auto.
    repeat rewrite associativity; eapply H0; eauto.
    intros; repeat rewrite associativity; eapply H1; eauto.
  Qed. *)

  End Exception_Bool.

(* Hint Extern 0 =>
  intros; match goal with
            |- (PAlgebra eval_Soundness_alg_Name
              (sig (UP'_P2 (eval_alg_Soundness_P _ _ _ _ _ _ _ _ _ _ _ _ _ _))) Arith) =>
            eapply Arith_eval_Soundness_alg; eauto with typeclass_instances
          end : typeclass_instances. *)


(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
