Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.
Require Import Ref.
Require Import Bool.

Section Ref_Bool.

  Variable F : Set -> Set.
  Context {Fun_F : Functor F}.
  Context {Sub_Except_F : RefE :<: F}.

  Definition Exp := Names.Exp F.

   Context {WF_Sub_Except_F : WF_Functor _ _ Sub_Except_F _ _}.

   Variable V : Set -> Set.
   Context {Fun_V : Functor V}.
   Definition Value := Value V.

   Context {Sub_StuckValue_V : StuckValue :<: V}.
   Definition stuck' : nat -> Value := stuck' _.
   Definition stuck : nat -> Fix V := stuck _.

   Context {Sub_LocValue_V : LocValue :<: V}.
   Context {Sub_BoolValue_V : BoolValue :<: V}.
   Context {Sub_UnitValue_V : UnitValue :<: V}.

  Variable (ME : Set -> Set). (* Evaluation Monad. *)
  Context {Fun_M : Functor ME}.
  Context {Mon_M : Monad ME}.
  Context {StateM_ME : StateM ME (list (Names.Value V))}.

  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.
  Context {Sub_RefType_D  : RefType :<: D}.
  Context {Sub_BType_D : BType :<: D}.
  Context {Sub_UnitType_D : UnitType :<: D}.

  Context {WF_Sub_UnitType_D : WF_Functor _ _ Sub_UnitType_D _ _}.
  Context {WF_Sub_AType_D : WF_Functor _ _ Sub_BType_D _ _}.
  Context {WF_Sub_RefType_D : WF_Functor _ _ Sub_RefType_D _ _}.

  Variable MT : Set -> Set.
  Context {Fun_MT : Functor MT}.
  Context {Mon_MT : Monad MT}.
  Context {Fail_MT : FailMonad MT}.

  Variable TypContext : Set.
  Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
  Context {funWFV : iFunctor WFV}.

  Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
  Variable funWFVM : iFunctor WFVM.

  Context {TypContextCE : ConsExtensionC TypContext}.
  Context {TypContext_WFE : WF_EnvC V TypContext}.

  Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME TypContext WFV) WFVM}.
  Context {Sub_WFVM_State_WFVM : Sub_iFunctor (WFValueM_State D V MT ME TypContext) WFVM}.

  Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) F}.
  Context {evalM_F : forall T, FAlgebra EvalName T (evalMR V ME) F}.
  Context {eq_DType_D : forall T : Set, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.

  Global Instance WFVM_State_eval_Soundness_Ifi :
    iPAlgebra Bool_eval_Soundness_Ifi_Name
    (Bool_eval_Soundness_Ifi_P D MT V ME _ WFVM) (WFValueM_State D V MT ME _).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
      try eassumption; unfold Bool_eval_Soundness_Ifi_P; split; simpl; intros.
    (* WFVM_Get *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor.
    intros env WF_env; apply (H0 env WF_env).
    rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM)).
    econstructor; intros; eapply H0; eauto.
    (* WFVM_Put *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor 2;
      eauto; eapply H2.
    unfold wbind; rewrite <- associativity.
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
    apply H2; eauto.
    intros; eapply H3; eapply ConsExtension_trans; eauto.
    intros; eapply H4; eapply ConsExtension_trans; eauto.
  Qed.

  Global Instance WFVM_State_eval_Soundness_Ift :
    iPAlgebra Bool_eval_Soundness_Ift_Name
    (Bool_eval_Soundness_Ift_P D MT V ME _ WFVM) (WFValueM_State D V MT ME _).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
      try eassumption; unfold Bool_eval_Soundness_Ift_P; split; simpl; intros.
    (* WFVM_Get *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor.
    intros env WF_env; apply (H0 env WF_env).
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)).
    econstructor; intros; eapply H0; eauto.
    (* WFVM_Put *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor 2;
      eauto; eapply H2.
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
    eapply (proj2 H2); eauto.
    intros; eapply H3; eapply ConsExtension_trans; eauto.
  Qed.

  Global Instance WFVM_State_eval_Soundness_Ife :
    iPAlgebra Bool_eval_Soundness_Ife_Name
    (Bool_eval_Soundness_Ife_P D MT V ME _ WFVM) (WFValueM_State D V MT ME _).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
      try eassumption; unfold Bool_eval_Soundness_Ife_P; split; simpl; intros.
    (* WFVM_Get *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor.
    intros env WF_env; apply (H0 env WF_env).
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)).
    econstructor; intros; eapply H0; eauto.
    (* WFVM_Put *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor 2;
      eauto; eapply H2.
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
    eapply (proj2 H2); eauto.
    intros; eapply H3; eapply ConsExtension_trans; eauto.
  Qed.

  Context {Dis_TUnit_BType : Distinct_Sub_Functor _ Sub_UnitType_D Sub_BType_D}.
  Context {Dis_TRef_BType : Distinct_Sub_Functor _ Sub_RefType_D Sub_BType_D}.

  Global Instance WF_invertVB_Unit :
    iPAlgebra WF_invertVB_Name (WF_invertVB_P D V _ WFV) (WFValue_Unit D V _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertVB_P; intros; inversion H; simpl in *.
    destruct i; injection H3; intros; subst; simpl in H0.
    rewrite H0 in H2.
    elimtype False.
    unfold tbool, tunit, tbool', tunit' in H2.
    generalize (inject_discriminate Dis_TUnit_BType
      (TUnit (sig Universal_Property'_fold)) (TBool (sig Universal_Property'_fold))).
    revert H2; unfold inject, inject'; eauto.
  Qed.

  Context {TypContext_S : SigTypContextC _ TypContext}.

  Global Instance WF_invertVB_Loc :
    iPAlgebra WF_invertVB_Name (WF_invertVB_P D V _ WFV) (WFValue_Loc D V _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertVB_P; intros; inversion H; simpl in *.
    destruct i; injection H4; intros; subst; simpl in H0.
    rewrite H0 in H3.
    elimtype False.
    unfold tbool, tref, tbool', tref' in H3.
    generalize (inject_discriminate Dis_TRef_BType
      (TRef (sig Universal_Property'_fold) T') (TBool (sig Universal_Property'_fold))).
    revert H3; unfold inject, inject'; destruct T'; eauto.
  Qed.

  Global Instance WF_invertLoc_VB :
   iPAlgebra WF_invertLoc_Name (WF_invertLoc_P D V _ WFV) (WFValue_VB D V _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertLoc_P; intros; inversion H; simpl in *.
    destruct i; injection H3; intros; subst; simpl in H0.
    rewrite H0 in H2.
    elimtype False.
    unfold tbool, tref, tbool', tref' in H2.
    generalize (inject_discriminate Dis_TRef_BType
      (TRef (sig Universal_Property'_fold) (exist _ T' T'_UP)) (TBool (sig Universal_Property'_fold))).
    revert H3; unfold inject, inject'; eauto.
  Qed.

  Context {Dis_LocValue_BoolValue : Distinct_Sub_Functor _ Sub_LocValue_V Sub_BoolValue_V}.
  Context {WF_Sub_LocValue_V : WF_Functor _ _ Sub_LocValue_V _ _}.
  Context {WF_Sub_BoolValue_V : WF_Functor _ _ Sub_BoolValue_V _ _}.

  Global Instance WF_invertLoc'_VB :
   iPAlgebra WF_invertLoc'_Name (WF_invertLoc'_P D V _ WFV) (WFValue_VB D V _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertLoc'_P; intros; inversion H; simpl in *.
    destruct i; injection H3; intros; subst; simpl in H0.
    rewrite H0 in H1.
    elimtype False.
    unfold loc, vb, loc', vb' in H1.
    generalize (inject_discriminate Dis_LocValue_BoolValue (Loc _ n) (VB _ n0)).
    revert H1; unfold inject, inject'; eauto.
  Qed.

  End Ref_Bool.

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
